"""GTAV 场景识别 - 基于小地图的楼层识别

核心思路：使用简单有效的特征 + KNN分类
"""

from pathlib import Path
from typing import Optional, Union
from dataclasses import dataclass, field
import cv2
import numpy as np
from tclogger import TCLogger
import pickle

logger = TCLogger("FloorRecognizer")

MODULE_DIR = Path(__file__).parents[2]
IMAGES_DIR = MODULE_DIR / "images"
CACHE_DIR = MODULE_DIR / "cache"


@dataclass
class MatchResult:
    """匹配结果"""

    floor: int  # 匹配到的楼层 (1, 2, 3)
    confidence: float  # 置信度 (0.0 - 1.0)
    score: float  # 匹配分数
    details: dict = field(default_factory=dict)


class FloorRecognizer:
    """基于小地图的楼层识别器。"""

    def __init__(
        self,
        floor_count: int = 3,
        images_dir: Optional[Path] = None,
        cache_dir: Optional[Path] = None,
    ):
        self.floor_count = floor_count
        self.images_dir = images_dir or IMAGES_DIR
        self.cache_dir = cache_dir or CACHE_DIR
        self.feature_db_path = self.cache_dir / "floor_features.pkl"

        # 各楼层的样本特征库 {floor: [(features, weight), ...]}
        self.floor_samples: dict[int, list[tuple[np.ndarray, float]]] = {
            i: [] for i in range(1, floor_count + 1)
        }

        # 加载地图特征
        self._load_floor_maps()

    def _load_floor_maps(self):
        """加载各楼层的完整平面图特征。"""
        for floor in range(1, self.floor_count + 1):
            map_path = self.images_dir / f"floor_{floor}_minimap.png"
            if not map_path.exists():
                map_path = self.images_dir / f"floor_{floor}_fullmap.png"
            if not map_path.exists():
                continue

            floor_map = cv2.imread(str(map_path), cv2.IMREAD_GRAYSCALE)
            if floor_map is not None:
                features = self._extract_features(floor_map)
                # 地图特征权重较低（因为和实际小地图差异大）
                self.floor_samples[floor].append((features, 0.5))
                logger.okay(f"楼层 {floor} 地图加载成功")

    def _extract_features(self, image: np.ndarray) -> np.ndarray:
        """提取用于楼层区分的特征。

        混合策略：局部特征（区分度高）+ 全局特征（泛化性好）
        """
        if len(image.shape) == 3:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        # 归一化尺寸
        image = cv2.resize(image, (64, 64))

        features = []

        # === 全局统计特征（泛化性好）===
        # 1. 全局灰度统计 (3)
        features.append(np.mean(image) / 255.0)
        features.append(np.std(image) / 128.0)
        features.append(np.median(image) / 255.0)

        # 2. 灰度直方图 (16 bins)
        hist = cv2.calcHist([image], [0], None, [16], [0, 256]).flatten()
        hist = hist / (hist.sum() + 1e-10)
        features.extend(hist)

        # === 边缘特征（结构性）===
        edges = cv2.Canny(image, 50, 150)

        # 3. 全局边缘密度 (1)
        features.append(np.mean(edges > 0))

        # 4. 边缘方向分布 (4) - 不同楼层可能有不同的建筑结构
        sobelx = cv2.Sobel(image, cv2.CV_64F, 1, 0, ksize=3)
        sobely = cv2.Sobel(image, cv2.CV_64F, 0, 1, ksize=3)
        magnitude = np.sqrt(sobelx**2 + sobely**2)

        # 水平边缘强度、垂直边缘强度
        features.append(np.mean(np.abs(sobelx)) / 255.0)
        features.append(np.mean(np.abs(sobely)) / 255.0)
        # 边缘强度标准差
        features.append(np.std(magnitude) / 255.0)

        # === 空间分布特征（3x3=9块，平衡局部和全局）===
        # 5. 分块均值 (9)
        block_size = 64 // 3
        for i in range(3):
            for j in range(3):
                y1, y2 = i * block_size, min((i + 1) * block_size, 64)
                x1, x2 = j * block_size, min((j + 1) * block_size, 64)
                block = image[y1:y2, x1:x2]
                features.append(np.mean(block) / 255.0)

        # 6. 分块边缘密度 (9)
        for i in range(3):
            for j in range(3):
                y1, y2 = i * block_size, min((i + 1) * block_size, 64)
                x1, x2 = j * block_size, min((j + 1) * block_size, 64)
                block = edges[y1:y2, x1:x2]
                features.append(np.mean(block > 0))

        # === 纹理特征 ===
        # 7. 拉普拉斯纹理复杂度 (1)
        laplacian = cv2.Laplacian(image, cv2.CV_64F)
        features.append(np.std(laplacian) / 100.0)

        # 8. 中心区域特征 (2) - 小地图中心通常是角色位置
        center = image[20:44, 20:44]
        features.append(np.mean(center) / 255.0)
        features.append(np.std(center) / 128.0)

        return np.array(features, dtype=np.float32)

    def _compute_distance(self, feat1: np.ndarray, feat2: np.ndarray) -> float:
        """计算欧氏距离。"""
        return np.linalg.norm(feat1 - feat2)

    def build_feature_database(self, save: bool = True):
        """从样本目录构建特征数据库。"""
        locations_dir = self.cache_dir / "locations"

        for floor in range(1, self.floor_count + 1):
            for d in locations_dir.glob(f"*floor_{floor}_minimap"):
                if not d.is_dir():
                    continue
                for file in d.glob("*.jpg"):
                    image = cv2.imread(str(file), cv2.IMREAD_GRAYSCALE)
                    if image is not None:
                        # 原始特征
                        features = self._extract_features(image)
                        self.floor_samples[floor].append((features, 1.0))

            count = len(self.floor_samples[floor])
            logger.note(f"楼层 {floor} 样本数：{count}")

        if save:
            self._save_database()

    def _save_database(self):
        """保存特征库。"""
        data = {
            floor: [(f.tolist(), w) for f, w in samples]
            for floor, samples in self.floor_samples.items()
        }
        self.feature_db_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.feature_db_path, "wb") as f:
            pickle.dump(data, f)
        logger.okay(f"特征库已保存: {self.feature_db_path}")

    def _load_database(self) -> bool:
        """加载特征库。"""
        if not self.feature_db_path.exists():
            return False
        try:
            with open(self.feature_db_path, "rb") as f:
                data = pickle.load(f)
            self.floor_samples = {
                int(k): [(np.array(f, dtype=np.float32), w) for f, w in v]
                for k, v in data.items()
            }
            logger.okay(f"特征库已加载")
            return True
        except Exception as e:
            logger.warn(f"加载失败: {e}")
            return False

    def recognize(self, image: np.ndarray) -> MatchResult:
        """识别楼层。使用加权KNN。"""
        features = self._extract_features(image)

        # 计算到所有样本的距离，按楼层分组
        floor_distances = {i: [] for i in range(1, self.floor_count + 1)}
        for floor, samples in self.floor_samples.items():
            for feat, weight in samples:
                dist = self._compute_distance(features, feat)
                floor_distances[floor].append((dist, weight))

        if all(len(d) == 0 for d in floor_distances.values()):
            return MatchResult(floor=1, confidence=0.0, score=0.0)

        # 对每个楼层排序，取每个楼层最近的k个样本
        k_per_floor = 3
        floor_scores = {i: 0.0 for i in range(1, self.floor_count + 1)}

        for floor, distances in floor_distances.items():
            if not distances:
                continue
            distances.sort(key=lambda x: x[0])
            top_k = distances[:k_per_floor]

            # 计算该楼层的得分（最近邻的距离越小，得分越高）
            for dist, weight in top_k:
                # 使用倒数权重
                dist_weight = 1.0 / (dist + 0.01)
                floor_scores[floor] += dist_weight * weight

        # 选择最佳
        best_floor = max(floor_scores.keys(), key=lambda f: floor_scores[f])
        best_score = floor_scores[best_floor]

        # 置信度
        sorted_scores = sorted(floor_scores.values(), reverse=True)
        if sorted_scores[0] > 0:
            confidence = (sorted_scores[0] - sorted_scores[1]) / sorted_scores[0]
        else:
            confidence = 0.0

        return MatchResult(
            floor=best_floor,
            confidence=confidence,
            score=best_score,
            details=floor_scores,
        )

    def recognize_file(self, path: Union[str, Path]) -> Optional[MatchResult]:
        """从文件识别。"""
        image = cv2.imread(str(path))
        if image is None:
            return None
        return self.recognize(image)

    def evaluate(self, test_dir: Path, expected_floor: int) -> dict:
        """评估准确率。"""
        files = list(Path(test_dir).glob("*.jpg"))
        if not files:
            return {"accuracy": 0, "total": 0, "correct": 0}

        correct = sum(
            1
            for f in files
            if (r := self.recognize_file(f)) and r.floor == expected_floor
        )
        return {
            "expected_floor": expected_floor,
            "total": len(files),
            "correct": correct,
            "accuracy": correct / len(files),
        }

    def __repr__(self) -> str:
        counts = {k: len(v) for k, v in self.floor_samples.items()}
        return f"FloorRecognizer(samples={counts})"


# ============================================================
# 测试
# ============================================================


def test_cross_validation():
    """留一法交叉验证 - 测试真实泛化能力。

    每次测试一个样本时，从特征库中排除该样本，
    这样可以避免"用训练数据测试自己"的过拟合假象。
    """
    logger.note("=" * 60)
    logger.note("留一法交叉验证（真实泛化能力测试）")
    logger.note("=" * 60)

    # 创建一个识别器用于特征提取（只创建一次）
    base_recognizer = FloorRecognizer()

    # 收集所有样本的特征
    locations_dir = CACHE_DIR / "locations"
    all_samples: dict[int, list[tuple[Path, np.ndarray]]] = {1: [], 2: [], 3: []}

    for floor in range(1, 4):
        for d in locations_dir.glob(f"*floor_{floor}_minimap"):
            if not d.is_dir():
                continue
            for f in d.glob("*.jpg"):
                image = cv2.imread(str(f), cv2.IMREAD_GRAYSCALE)
                if image is not None:
                    features = base_recognizer._extract_features(image)
                    all_samples[floor].append((f, features))

    for floor, samples in all_samples.items():
        logger.note(f"楼层 {floor} 样本数：{len(samples)}")

    # 留一法交叉验证
    results = {
        1: {"correct": 0, "total": 0},
        2: {"correct": 0, "total": 0},
        3: {"correct": 0, "total": 0},
    }
    total_correct = 0
    total_samples = 0

    for test_floor in range(1, 4):
        for i, (test_file, test_features) in enumerate(all_samples[test_floor]):
            # 清空样本库，重新构建（排除当前测试样本）
            base_recognizer.floor_samples = {f: [] for f in range(1, 4)}

            # 添加所有其他样本
            for floor in range(1, 4):
                for j, (f, features) in enumerate(all_samples[floor]):
                    if floor == test_floor and j == i:
                        continue  # 排除测试样本
                    base_recognizer.floor_samples[floor].append((features, 1.0))

            # 直接用已提取的特征进行识别（避免重复读取图像）
            # 模拟 recognize 的逻辑
            all_distances = []
            # 按楼层分组计算距离
            floor_distances = {f: [] for f in range(1, 4)}
            for floor, samples in base_recognizer.floor_samples.items():
                for feat, weight in samples:
                    dist = base_recognizer._compute_distance(test_features, feat)
                    floor_distances[floor].append((dist, weight))

            # 每个楼层取最近的k个样本
            k_per_floor = 3
            floor_scores = {f: 0.0 for f in range(1, 4)}

            for floor, distances in floor_distances.items():
                if not distances:
                    continue
                distances.sort(key=lambda x: x[0])
                top_k = distances[:k_per_floor]
                for dist, weight in top_k:
                    dist_weight = 1.0 / (dist + 0.01)
                    floor_scores[floor] += dist_weight * weight

            predicted_floor = max(floor_scores.keys(), key=lambda f: floor_scores[f])

            results[test_floor]["total"] += 1
            if predicted_floor == test_floor:
                results[test_floor]["correct"] += 1

        # 每完成一个楼层输出进度
        r = results[test_floor]
        acc = r["correct"] / r["total"] if r["total"] > 0 else 0
        logger.mesg(f"楼层 {test_floor}: {acc:.1%} ({r['correct']}/{r['total']})")
        total_correct += r["correct"]
        total_samples += r["total"]

    overall = total_correct / total_samples if total_samples > 0 else 0
    logger.okay(f"总体泛化准确率: {overall:.1%} ({total_correct}/{total_samples})")


def test_accuracy():
    """测试准确率（注意：这是用训练数据测试，可能过拟合）。"""
    logger.note("=" * 60)
    logger.note("准确率测试（训练集自测，可能过拟合）")
    logger.note("=" * 60)

    recognizer = FloorRecognizer()

    if not recognizer._load_database():
        recognizer.build_feature_database(save=True)

    logger.note(f"{recognizer}")

    locations_dir = CACHE_DIR / "locations"
    results = []

    for d in sorted(locations_dir.glob("*_minimap")):
        for floor in range(1, 4):
            if f"floor_{floor}" in d.name:
                r = recognizer.evaluate(d, floor)
                results.append(r)
                logger.mesg(
                    f"{d.name}: {r['accuracy']:.1%} ({r['correct']}/{r['total']})"
                )
                break

    if results:
        total = sum(r["total"] for r in results)
        correct = sum(r["correct"] for r in results)
        logger.okay(f"总体: {correct/total:.1%} ({correct}/{total})")


def test_speed():
    """测试速度。"""
    logger.note("=" * 60)
    logger.note("速度测试")
    logger.note("=" * 60)

    recognizer = FloorRecognizer()
    if not recognizer._load_database():
        recognizer.build_feature_database(save=True)

    # 收集测试图像
    test_files = []
    for d in (CACHE_DIR / "locations").glob("*_minimap"):
        test_files.extend(list(d.glob("*.jpg"))[:5])

    if not test_files:
        return

    import time

    # 预热
    recognizer.recognize_file(test_files[0])

    start = time.time()
    for f in test_files:
        recognizer.recognize_file(f)
    elapsed = time.time() - start

    logger.okay(
        f"平均: {elapsed/len(test_files)*1000:.1f} ms/张 ({len(test_files)} 张)"
    )


if __name__ == "__main__":
    test_speed()
    test_accuracy()  # 训练集自测（可能过拟合）
    test_cross_validation()  # 留一法交叉验证 - 测试真实泛化能力

    # python -m gtaz.models.agency_move.recognizes



"""GTAV 菜单导航模块"""

from tclogger import TCLogger, logstr
from typing import Union

from .locates import MatchResult, MergedMatchResult
from .commons import MENU_FOCUS_INFOS, MENU_PARENT_TO_ITEM_INFOS
from .commons import STORY_MENU_FOCUS_INFOS, STORY_MENU_PARENT_TO_ITEM_INFOS
from .commons import is_names_start_with
from .interacts import MenuInteractor
from .locates import MenuLocatorRunner, is_score_too_low, is_score_high
from ..visions.screens import ScreenCapturer

logger = TCLogger(name="MenuNavigator", use_prefix=True, use_prefix_ms=True)

ItemType = Union[str, int]
ActionType = tuple[str, int]


class MenuInfoProvider:
    """菜单信息提供器，根据模式返回对应的菜单信息"""

    def __init__(self, netmode: str = None):
        """初始化菜单信息提供器

        :param netmode: 模式名称，"界面模式"/"在线模式"/"故事模式"
        """
        self._netmode = netmode
        self._update_infos()

    def _update_infos(self):
        """根据当前模式更新菜单信息"""
        if self._netmode == "故事模式":
            self.focus_infos = STORY_MENU_FOCUS_INFOS
            self.parent_to_item_infos = STORY_MENU_PARENT_TO_ITEM_INFOS
        elif self._netmode == "在线模式":
            self.focus_infos = MENU_FOCUS_INFOS
            self.parent_to_item_infos = MENU_PARENT_TO_ITEM_INFOS

    @property
    def netmode(self) -> str:
        """获取当前模式"""
        return self._netmode

    @netmode.setter
    def netmode(self, value: str):
        """设置模式并更新菜单信息

        :param value: 模式名称，"界面模式"/"在线模式"/"故事模式"
        """
        if self._netmode != value:
            self._netmode = value
            self._update_infos()

    def unify_tab(self, tab: ItemType) -> tuple[int, str]:
        """统一标签页表示形式

        :param tab: 标签页名称或索引

        :return: (标签页索引, 标签页名称)
        """
        tabs_name_idxs = {info["name"]: info["index"] for info in self.focus_infos}
        tabs_idx_names = {info["index"]: info["name"] for info in self.focus_infos}
        if isinstance(tab, int):
            tab_idx = tab
            tab_name = tabs_idx_names.get(tab_idx)
        else:
            tab_name = tab
            tab_idx = tabs_name_idxs.get(tab_name)
        return tab_idx, tab_name

    @staticmethod
    def unify_item(item: ItemType, item_infos: list[dict]) -> tuple[int, str]:
        """统一菜单项表示形式

        :param item: 菜单项名称或索引
        :param item_infos: 该级菜单项信息列表

        :return: (菜单项索引, 菜单项名称)
        """
        item_idx_names = {info["index"]: info["name"] for info in item_infos}
        item_name_idxs = {info["name"]: info["index"] for info in item_infos}
        if isinstance(item, int):
            item_idx = item
            item_name = item_idx_names.get(item_idx)
        else:
            item_name = item
            item_idx = item_name_idxs.get(item_name)
        return item_idx, item_name


class LocateNamesConverter:
    @staticmethod
    def _is_list_belongs_to_focus(r: MergedMatchResult) -> bool:
        """判断 focus.names 是否是 list.names 的前缀"""
        return is_names_start_with(r.list.names, r.focus.names)

    @staticmethod
    def _is_item_belongs_to_list(r: MergedMatchResult) -> bool:
        """判断 list.names 是否是 item.names 的前缀"""
        return is_names_start_with(r.item.names, r.list.names)

    @staticmethod
    def _is_item_belongs_to_focus(r: MergedMatchResult) -> bool:
        """判断focus.names 是否是 item.names 的前缀"""
        return is_names_start_with(r.item.names, r.focus.names)

    def to_names(self, r: MergedMatchResult) -> list[str]:
        """将菜单定位结果转换为菜单路径表示形式"""
        names = []

        # header匹配结果仅用于判断菜单是否打开
        if is_score_too_low(r.header):
            # 菜单未打开，返回空路径
            return []

        # focus匹配结果用于获取当前菜单标题
        if is_score_too_low(r.focus):
            # 菜单打开但无法定位标题，或者菜单未打开，返回空路径
            return []
        else:
            # 菜单打开且能定位标题
            names.append(r.focus.name)

        # list匹配结果用于获取当前菜单列表
        if is_score_too_low(r.list) or not self._is_list_belongs_to_focus(r):
            # 无法匹配已知列表，或不属于当前标题
            # 通常是刚切换到标题，列表还未激活或加载完成
            # 暂时继续往下尝试匹配条目
            pass

        # item匹配结果用于获取当前菜单条目
        if (
            is_score_too_low(r.item)
            or not self._is_item_belongs_to_list(r)
            or not self._is_item_belongs_to_focus(r)
        ):
            # 无法匹配已知条目，或不属于当前列表
            return names
        else:
            # 使用 item.names 作为最终路径
            return list(r.item.names)

        return names


class Action:
    """菜单导航动作"""

    TOGGLE_MENU = "toggle_menu"
    CONFIRM = "confirm"
    CANCEL = "cancel"
    NAV_UP = "nav_up"
    NAV_DOWN = "nav_down"
    TAB_LEFT = "tab_left"
    TAB_RIGHT = "tab_right"


class MenuNavigatePlanner:
    """GTAV 菜单导航规划"""

    def __init__(self, netmode: str = None):
        """初始化导航规划器

        :param netmode: 模式名称，"在线模式" 或 "故事模式"
        """
        self.info_provider = MenuInfoProvider(netmode)

    @property
    def netmode(self) -> str:
        """获取当前模式"""
        return self.info_provider.netmode

    @netmode.setter
    def netmode(self, value: str):
        """设置模式

        :param value: 模式名称，"在线模式" 或 "故事模式"
        """
        self.info_provider.netmode = value

    def _calc_tab_switch_action(
        self, src_tab: ItemType, dst_tab: ItemType
    ) -> ActionType:
        """计算标签页切换的动作

        :param src_tab: 当前标签页名称
        :param dst_tab: 目标标签页名称

        :return: 动作，格式为 (action, times) 元组
        """
        src_tab_idx, _ = self.info_provider.unify_tab(src_tab)
        dst_tab_idx, _ = self.info_provider.unify_tab(dst_tab)

        if src_tab_idx is None or dst_tab_idx is None:
            logger.warn(f"无法识别标签输入: {src_tab} 或 {dst_tab}")
            return None
        # 当前标签页即目标标签页
        if src_tab_idx == dst_tab_idx:
            return (None, 0)
        # 计算移动次数
        tabs_num = len(self.info_provider.focus_infos)
        right_moves = (dst_tab_idx - src_tab_idx) % tabs_num
        left_moves = (src_tab_idx - dst_tab_idx) % tabs_num
        if right_moves <= left_moves:
            actions = (Action.TAB_RIGHT, right_moves)
        else:
            actions = (Action.TAB_LEFT, left_moves)
        return actions

    def _calc_item_nav_action(
        self, src_item: ItemType, dst_item: ItemType, item_infos: list[dict]
    ) -> ActionType:
        """
        计算菜单项导航的动作

        :param src_item: 当前菜单项名称或索引
        :param dst_item: 目标菜单项名称或索引
        :param item_infos: 该级菜单项信息列表
        :return: 动作，格式为 (action, times) 元组
        """
        src_item_idx, _ = self.info_provider.unify_item(src_item, item_infos)
        dst_item_idx, _ = self.info_provider.unify_item(dst_item, item_infos)
        items_num = len(item_infos)
        if src_item_idx is None or dst_item_idx is None:
            logger.warn(f"无法识别菜单项输入: {src_item} 或 {dst_item}")
            return None
        # 当前菜单项即目标菜单项
        if src_item_idx == dst_item_idx:
            return (None, 0)
        # 计算移动次数
        down_moves = (dst_item_idx - src_item_idx) % items_num
        up_moves = (src_item_idx - dst_item_idx) % items_num
        if down_moves - 4 <= up_moves:
            # 向下移动不容易卡住，因此多留点余地
            # 例如在故事模式，当向上移动越过位置0时，理应跳到最底部的菜单项，但经常会卡住
            actions = (Action.NAV_DOWN, down_moves)
        else:
            actions = (Action.NAV_UP, up_moves)
        return actions

    def sum_actions_times(self, actions: list[ActionType]) -> int:
        """计算动作列表中所有动作的总次数"""
        total = 0
        for action, times in actions:
            if action and times:
                total += times
        return total

    @staticmethod
    def _log_error_parent_names(parent_names: list[str]) -> None:
        logger.warn(f"无法获取菜单项信息: 父级菜单路径 {parent_names}")

    def _calc_common_prefix_length(
        self, src_names: list[str], dst_names: list[str]
    ) -> int:
        """找到两个路径的公共前缀长度

        :param src_names: 当前菜单项名称列表
        :param dst_names: 目标菜单项名称列表

        :return: 公共前缀长度
        """
        common_len = 0
        min_len = min(len(src_names), len(dst_names))
        for i in range(min_len):
            if src_names[i] == dst_names[i]:
                common_len += 1
            else:
                break
        return common_len

    def _plan_backward(
        self, src_names: list[str], dst_names: list[str]
    ) -> list[ActionType]:
        """规划回退到目标路径

        :param src_names: 当前菜单项名称列表
        :param dst_names: 公共前缀名称列表

        :return: 回退动作列表
        """
        actions: list[ActionType] = []
        # 回退到 common_len 层级
        cancel_times = len(src_names) - len(dst_names)
        if cancel_times > 0:
            actions.append((Action.CANCEL, cancel_times))
        return actions

    def _plan_backward_to_sibling(
        self, src_names: list[str], dst_names: list[str]
    ) -> list[ActionType]:
        """规划回退到平级位置

        :param src_names: 当前菜单项名称列表
        :param dst_names: 公共前缀名称列表

        :return: 回退动作列表
        """
        actions: list[ActionType] = []
        # 回退到 common_len + 1 层级，也即平级位置
        cancel_times = len(src_names) - len(dst_names) - 1
        if cancel_times > 0:
            actions.append((Action.CANCEL, cancel_times))
        return actions

    def _plan_forward(
        self, src_names: list[str], dst_names: list[str]
    ) -> list[ActionType]:
        """规划从公共前缀位置前进到目标位置

        :param src_names: 公共前缀名称列表
        :param dst_names: 目标菜单项名称列表

        :return: 前进动作列表
        """
        actions: list[ActionType] = []
        if src_names == dst_names:
            return []
        # 从公共前缀的下一级开始，逐级前进到目标
        for level in range(len(src_names) + 1, len(dst_names) + 1):
            # 确认进入下一级菜单
            actions.append((Action.CONFIRM, 1))
            # 获取当前层级的父级路径和映射信息
            parent_names = dst_names[: level - 1]
            item_infos = self.info_provider.parent_to_item_infos.get(
                tuple(parent_names)
            )
            if item_infos is None:
                self._log_error_parent_names(parent_names)
                break
            # 导航到目标条目
            dst_name = dst_names[level - 1]
            nav_action = self._calc_item_nav_action(0, dst_name, item_infos=item_infos)
            actions.append(nav_action)
        return actions

    def _is_sibling(self, src_names: list[str], dst_names: list[str]) -> bool:
        """判断当前路径和目标路径是否平级

        也即：路径长度相同，且有公共父级，只有最后一项不同
        例如: ['在线', '差事'] -> ['在线', '游玩清单']

        :param src_names: 当前菜单项名称列表
        :param dst_names: 目标菜单项名称列表
        :param common_len: 公共前缀长度

        :return: 是否为平级切换
        """
        common_len = self._calc_common_prefix_length(src_names, dst_names)
        return (
            len(src_names) == len(dst_names)  # 路径长度相同
            and len(src_names) > 1  # 至少有一级父级
            and common_len == len(src_names) - 1  # 只有最后一项不同
        )

    def _plan_sibling(
        self, src_names: list[str], dst_names: list[str]
    ) -> list[ActionType]:
        """规划平级切换的导航路径

        :param src_names: 当前菜单项名称列表
        :param dst_names: 目标菜单项名称列表

        :return: 导航路径动作列表
        """
        actions: list[ActionType] = []
        if src_names == dst_names:
            return []
        # 获取父级路径
        parent_names = src_names[:-1]
        # 获取该级的条目信息
        item_infos = self.info_provider.parent_to_item_infos.get(tuple(parent_names))
        if item_infos is None:
            logger.warn(f"无法获取菜单项信息: 父级菜单路径 {parent_names}")
            return actions
        # 获取当前和目标条目名称
        src_item_name = src_names[-1]
        dst_item_name = dst_names[-1]
        # 计算从当前项到目标项的导航动作
        nav_action = self._calc_item_nav_action(
            src_item_name, dst_item_name, item_infos=item_infos
        )
        actions.append(nav_action)
        return actions

    def plan_from_origin(self, dst_names: list[str]) -> list[ActionType]:
        """从头规划导航路径：从菜单关闭状态，到 names 对应位置

        :param dst_names: 目标菜单项名称列表

        :return: 导航路径动作列表，每个动作为 (action, times) 元组
        """
        actions: list[ActionType] = []
        if not dst_names:
            return []
        # 打开菜单
        actions.append((Action.TOGGLE_MENU, 1))
        # 切换到目标标签页
        dst_tab_name = dst_names[0]
        tab_action = self._calc_tab_switch_action(0, dst_tab_name)
        actions.append(tab_action)
        # 前进到目标位置
        forward_actions = self._plan_forward([dst_tab_name], dst_names)
        actions.extend(forward_actions)
        return actions

    def plan_from_source(
        self, src_names: list[str], dst_names: list[str]
    ) -> list[ActionType]:
        """规划导航路径: 从 src_names 到 dst_names

        :param src_names: 当前菜单项名称列表
        :param dst_names: 目标菜单项名称列表

        :return: 导航路径动作列表，每个动作为 (action, times) 元组
        """
        # 边界情况: 当前菜单未打开，从头规划
        if not src_names:
            return self.plan_from_origin(dst_names)
        # 边界情况: 目标是关闭菜单，直接关闭菜单
        if not dst_names:
            return [(Action.TOGGLE_MENU, 1)]
        # 边界情况: 当前位置即目标位置，什么都不做
        if src_names == dst_names:
            return []

        actions: list[ActionType] = []

        # 找到公共前缀
        common_len = self._calc_common_prefix_length(src_names, dst_names)
        common_names = src_names[:common_len]

        # 特殊情况：跨标签页
        if common_len == 0:
            tab_action = self._calc_tab_switch_action(src_names[0], dst_names[0])
            actions.append(tab_action)
            # 从标签页切换后，前进到目标位置
            forward_actions = self._plan_forward([dst_names[0]], dst_names)
            actions.extend(forward_actions)
            return actions

        # 从当前位置回退到公共前缀下一级，也即平级位置
        if len(src_names) >= common_len + 1:
            backward_actions = self._plan_backward_to_sibling(src_names, common_names)
            actions.extend(backward_actions)

        # 使用公共前缀下一级，作为中间位置
        mid_names = src_names[: common_len + 1]
        mid_len = len(mid_names)

        # 目标恰好为公共前缀，回退
        if len(dst_names) < mid_len:
            # 事实上等价于 len(dst_names) == common_len
            backward_actions = self._plan_backward(mid_names, dst_names)
            actions.extend(backward_actions)

        # 目标在平级或者更深层级，平移
        if len(dst_names) >= mid_len:
            sibling_actions = self._plan_sibling(mid_names, dst_names[:mid_len])
            actions.extend(sibling_actions)

        # 目标为中间位置的更深层级，前进
        if len(dst_names) > mid_len:
            forward_actions = self._plan_forward(mid_names, dst_names)
            actions.extend(forward_actions)

        return actions


class MenuNavigator:
    def __init__(self, verbose: bool = False):
        self.interactor = MenuInteractor()
        self.locator_runner = MenuLocatorRunner()
        self.capturer = ScreenCapturer()
        self.converter = LocateNamesConverter()
        self.planner = MenuNavigatePlanner()
        self.netmode: str = None
        self.verbose = verbose

    def _update_netmode_from_result(self, result: MergedMatchResult) -> None:
        """根据定位结果更新当前网络模式

        :param result: 定位结果
        """
        if result and result.netmode and not is_score_too_low(result.netmode):
            netmode_name = result.netmode.name
            if self.netmode != netmode_name:
                self.netmode = netmode_name
                self.planner.netmode = netmode_name
                logger.mesg(f"导航模式切换到: [{netmode_name}]")

    def locate(self) -> MergedMatchResult:
        """获取当前菜单定位结果"""
        try:
            frame_np = self.capturer.capture_frame(verbose=self.verbose).to_np()
            result = self.locator_runner.locate(frame_np, verbose=self.verbose)
            self._update_netmode_from_result(result)
        except Exception as e:
            logger.erro(e)
            result = MergedMatchResult()
        return result

    def ensure_menu_opened(self, max_retries: int = 5):
        """确保菜单处于打开状态"""
        retry = 0
        while retry < max_retries:
            retry += 1
            result = self.locate()
            if is_score_too_low(result.netmode):
                # 菜单未打开，执行打开操作
                # 对于故事模式，需要 toggle_menu 两次
                self.interactor.toggle_menu()
            else:
                # 菜单已打开
                break

    def ensure_menu_closed(self, max_retries: int = 5):
        """确保菜单处于关闭状态"""
        retry = 0
        while retry < max_retries:
            retry += 1
            result = self.locate()
            if not is_score_too_low(result.netmode):
                # 菜单未关闭，执行关闭操作
                # 取消5次：
                # - 在故事和在线模式中，可以退出绝大多数层级；
                # - 在界面模式中，恰好可以关闭退出提示
                self.interactor.cancel(5)
            else:
                # 菜单已关闭
                break

    def locate_names(self) -> list[str]:
        """获取当前菜单项名称列表"""
        locate_result = self.locate()
        return self.converter.to_names(locate_result)

    def plan_actions(self, dst_names: list[str]) -> list[ActionType]:
        """规划导航到指定菜单项的动作"""
        self.ensure_menu_opened()
        src_names = self.locate_names()
        actions = self.planner.plan_from_source(src_names, dst_names)
        return actions

    def execute_actions(self, actions: list[ActionType]):
        """执行导航动作列表

        :param actions: 动作列表，每项为 (action, times) 元组
        """
        for action in actions:
            if not action:
                continue
            act, times = action
            if not act or not times:
                continue
            elif act == Action.TOGGLE_MENU:
                self.interactor.toggle_menu()
            elif act == Action.CONFIRM:
                self.interactor.confirm()
            elif act == Action.CANCEL:
                self.interactor.cancel(times)
            elif act == Action.NAV_UP:
                self.interactor.nav_up(times)
            elif act == Action.NAV_DOWN:
                self.interactor.nav_down(times)
            elif act == Action.TAB_LEFT:
                self.interactor.tab_left(times)
            elif act == Action.TAB_RIGHT:
                self.interactor.tab_right(times)
            else:
                logger.warn(f"未知导航动作: {act}")
            # 每次操作后等待菜单稳定
            self.interactor.wait_until_ready()

    def _log_retry_go_to(self, names: list[str], retry: int, max_retries: int) -> None:
        """打印导航重试日志"""
        # 第一次不打印日志，避免冗余信息
        if retry >= 2:
            logger.warn(f"[{retry}/{max_retries}] 未导航到预期位置，当前路径: {names}")

    def go_to(self, dst_names: list[str], max_retries: int = 5) -> list[str]:
        """导航到指定菜单项

        :param dst_names: 目标菜单项名称列表
        :param max_retries: 最大重试次数

        :return: 当前菜单项名称列表
        """
        retry = 0
        while retry < max_retries:
            actions = self.plan_actions(dst_names)
            if self.verbose:
                logger.mesg(f"当前模式: [{self.planner.netmode}]")
                logger.mesg(f"导航动作: {logstr.file(actions)}")
            self.execute_actions(actions)
            names = self.locate_names()
            if names == dst_names:
                return names
            else:
                retry += 1
                self._log_retry_go_to(names, retry, max_retries)
                continue
        return names


def test_planner():
    logger.note("测试: MenuNavigatePlanner...")
    planner = MenuNavigatePlanner()

    def _log_plan_from_origin(names: list[str]):
        logger.mesg(f"规划路径 (从原点): {names}")
        paths = planner.plan_from_origin(names)
        logger.okay(f"规划结果: {paths}")
        logger.mesg(f"总步数: {planner.sum_actions_times(paths)}\n")

    def _log_plan_from_source(src_names: list[str], dst_names: list[str]):
        logger.mesg(f"规划路径: {src_names} -> {dst_names}")
        paths = planner.plan_from_source(src_names, dst_names)
        logger.okay(f"规划结果: {paths}")
        logger.mesg(f"总步数: {planner.sum_actions_times(paths)}\n")

    # 测试从原点出发的路径规划
    logger.note("=== 测试: 从原点出发 ===")
    origin_cases = [
        ["设置"],
        ["在线", "差事", "进行差事"],
        ["在线", "差事", "进行差事", "已收藏的"],
    ]
    for names in origin_cases:
        _log_plan_from_origin(names)

    # 测试完整路径规划
    logger.note("=== 测试: 边界情况 ===")
    boundary_cases = [
        ([], ["在线"]),  # 空路径到目标
        (["在线"], []),  # 当前位置到空路径 (关闭菜单)
        (["在线"], ["在线"]),  # 当前位置即目标位置
    ]
    for src, dst in boundary_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 同级标签页切换 ===")
    tab_switch_cases = [
        (["在线"], ["设置"]),
        (["设置"], ["地图"]),
    ]
    for src, dst in tab_switch_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 同一标签页内的导航 ===")
    same_tab_nav_cases = [
        (["在线"], ["在线", "差事"]),  # 一级到二级
        (["在线"], ["在线", "退出游戏"]),  # 一级到二级
        (["在线", "差事"], ["在线", "差事", "进行差事"]),  # 二级到三级
        (
            ["在线", "差事", "进行差事"],
            ["在线", "差事", "进行差事", "已收藏的"],
        ),  # 三级到四级
        (["在线"], ["在线", "差事", "进行差事", "已收藏的"]),  # 一级到四级
    ]
    for src, dst in same_tab_nav_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 同一标签页内的回退 ===")
    backward_cases = [
        (
            ["在线", "差事", "进行差事", "已收藏的"],
            ["在线", "差事", "进行差事"],
        ),  # 四级到三级
        (["在线", "差事", "进行差事"], ["在线", "差事"]),  # 三级到二级
        (["在线", "差事", "进行差事", "已收藏的"], ["在线"]),  # 四级到一级
    ]
    for src, dst in backward_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 同一标签页内的平级切换 ===")
    sibling_switch_cases = [
        (["在线", "差事"], ["在线", "游玩清单"]),  # 优化: 只需要导航
        (
            ["在线", "差事", "快速加入"],
            ["在线", "差事", "进行差事"],
        ),  # 优化: 只需要导航
    ]
    for src, dst in sibling_switch_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 跨标签页的复杂导航 ===")
    cross_tab_cases = [
        (["在线", "差事", "进行差事", "已收藏的"], ["设置"]),  # 优化: 无需回退
        (["设置"], ["在线", "差事", "进行差事", "已收藏的"]),  # 优化: 无需回退
        (["在线", "差事", "进行差事"], ["地图"]),  # 优化: 无需回退
    ]
    for src, dst in cross_tab_cases:
        _log_plan_from_source(src, dst)

    logger.note("=== 测试: 不同分支的切换 ===")
    branch_switch_cases = [
        (
            ["在线", "差事", "进行差事"],
            ["在线", "寻找新战局"],
        ),
        (
            ["在线", "差事", "进行差事", "已收藏的"],
            ["在线", "游玩清单"],
        ),
        (
            ["在线", "加入帮会成员"],
            ["在线", "差事", "进行差事", "已收藏的"],
        ),
    ]
    for src, dst in branch_switch_cases:
        _log_plan_from_source(src, dst)


def test_menu_navigator():
    logger.note("测试: MenuNavigator...")
    navigator = MenuNavigator()
    names = navigator.locate_names()
    logger.mesg(f"当前路径: {names}")

    # dst_names = ["在线", "差事"]
    # dst_names = ["在线", "差事", "进行差事", "已收藏的", "夺取"]
    # dst_names = ["在线", "进入GTA在线模式"]
    dst_names = ["在线", "进入GTA在线模式", "凭邀请加入的战局"]
    # dst_names = ["游戏", "退至主菜单"]
    logger.mesg(f"导航到: {dst_names}")
    names = navigator.go_to(dst_names)
    logger.mesg(f"当前路径: {names}")


if __name__ == "__main__":
    # test_planner()
    test_menu_navigator()

    # python -m gtaz.menus.navigates



"""GTA 在线/故事 模式切换模块"""

import argparse
import sys
import time

from tclogger import TCLogger, logstr
from acto import SoftRetrier
from typing import Union

from ..menus.navigates import MenuNavigator
from ..menus.locates import ExitLocatorRunner, is_score_too_low
from ..visions.screens import ScreenCapturer


logger = TCLogger(name="NetmodeSwitcher", use_prefix=True, use_prefix_ms=True)


class NetmodeSwitcher:
    """GTA 界面/在线/故事模式切换器"""

    def __init__(self):
        """初始化模式切换器"""
        self.navigator = MenuNavigator()
        self.capturer = ScreenCapturer()
        # 使用 navigator 中的 interactor，避免创建新的手柄（会不生效）
        self.interactor = self.navigator.interactor
        # 使用 navigator 中的 locator
        self.locator = self.navigator.locator_runner.locator
        self.exit_runner = ExitLocatorRunner()

    def get_netmode(self) -> Union[str, bool]:
        """获取当前模式

        :return: 当前模式名称，"界面模式"/"在线模式"/"故事模式"，如果无法识别则返回 None
        """
        # 确保菜单打开
        self.navigator.ensure_menu_opened()
        # 截取屏幕
        frame_np = self.capturer.capture_frame(verbose=False).to_np()
        # 匹配模式
        result = self.locator.match_mode(frame_np)
        # 判断匹配结果
        if not is_score_too_low(result):
            mode_name = result.name
            logger.mesg(f"当前模式: [{mode_name}]")
            # # 对于界面模式，关闭菜单，以进行后续切换
            # if mode_name == "界面模式":
            #     self.navigator.ensure_menu_closed()
            #     logger.mesg("已关闭设置菜单，回到开始页")
            return mode_name
        else:
            logger.warn("无法识别当前模式")
            return False

    def get_netmode_with_retry(self):
        retrier = SoftRetrier(max_retries=10, retry_interval=5)
        mode = retrier.run(self.get_netmode)
        return mode

    def exit_and_confirm(self, max_retries: int = 10) -> bool:
        """定位退出提示并确认

        :param max_retries: 最大重试次数
        :return: 是否成功完成定位-确认流程
        """
        logger.note("定位退出提示并确认...")

        retry = 0
        while retry < max_retries:
            retry += 1
            # 等待退出提示出现
            self.interactor.wait_until_ready(500)
            # 截取屏幕
            frame_np = self.capturer.capture_frame(verbose=False).to_np()
            # 定位退出提示
            exit_result = self.exit_runner.locate(frame_np, verbose=True)
            # 判断是否匹配到退出提示
            if not is_score_too_low(exit_result):
                logger.mesg(f"已看到退出提示: {exit_result.name}")
                # 确认退出
                self.interactor.confirm()
                logger.okay("已确认退出提示")
                return True
            else:
                retry_str = f"[{retry} / {max_retries}]"
                logger.warn(f"{retry_str} 未能看到退出提示，重试中...")
                # 再按一次 confirm，可能需要多次确认才能触发退出提示
                self.interactor.confirm()
        logger.fail("未能看到退出提示")
        return False

    def _switch_mode(
        self, dst_names: list[str], mode_desc: str, max_retries: int = 5
    ) -> bool:
        """通用模式切换方法

        :param dst_names: 目标菜单路径
        :param mode_desc: 模式切换描述（用于日志）
        :param max_retries: 最大重试次数
        :return: 是否成功切换模式
        """
        logger.note("=" * 50)
        logger.note(mode_desc)
        logger.note("=" * 50)

        retry = 0
        while retry < max_retries:
            retry += 1
            if retry > 1:
                retry_str = f"[{retry}/{max_retries}] "
            else:
                retry_str = ""
            logger.note(f"{retry_str}尝试导航到: {dst_names}")
            # 导航到目标菜单项
            current_names = self.navigator.go_to(dst_names)
            # 检查是否导航到目标位置
            if current_names == dst_names:
                logger.okay(f"成功导航到: {current_names}")
                # 确认选择
                self.interactor.confirm()
                logger.mesg("已确认选择，等待退出提示...")
                # 定位退出提示并确认
                if self.exit_and_confirm():
                    # logger.okay("已完成切换")
                    return True
                # 如果定位退出提示失败，继续重试整个流程
            else:
                logger.warn(f"未导航到目标位置，当前位置: {current_names}，重试...")
                # 关闭菜单并重新打开
                self.navigator.ensure_menu_closed()
                self.navigator.ensure_menu_opened()

        logger.fail("模式切换失败")
        return False

    def _log_okay_mode(self, mode_name: str) -> None:
        """记录当前模式

        :param mode_name: 当前模式名称
        """
        logger.note(f"当前已是{logstr.okay(mode_name)}，无需切换")

    def switch_story_to_online(self, max_retries: int = 5) -> bool:
        """故事模式切换到在线模式

        :param max_retries: 最大重试次数
        :return: 是否成功切换模式
        """
        # 检查当前模式
        mode = self.get_netmode_with_retry()
        if mode == "在线模式":
            self._log_okay_mode(mode)
            return True

        dst_names = ["在线", "进入GTA在线模式", "凭邀请加入的战局"]
        return self._switch_mode(dst_names, "故事模式 -> 在线模式", max_retries)

    def switch_online_to_story(self, max_retries: int = 5) -> bool:
        """在线模式切换到故事模式

        :param max_retries: 最大重试次数
        :return: 是否成功切换模式
        """
        # 检查当前模式
        mode = self.get_netmode_with_retry()
        if mode == "故事模式":
            self._log_okay_mode(mode)
            return True

        dst_names = ["在线", "退至故事模式"]
        return self._switch_mode(dst_names, "在线模式 -> 故事模式", max_retries)

    def switch_to_new_invite_lobby(self, max_retries: int = 5) -> bool:
        """切换到新的邀请战局

        根据当前模式自动选择对应的菜单路径：
        - 故事模式: 在线 -> 进入GTA在线模式 -> 凭邀请加入的战局
        - 在线模式: 在线 -> 寻找新战局 -> 仅限邀请的战局

        :param max_retries: 最大重试次数
        :return: 是否成功切换到新的邀请战局
        """
        # 获取当前模式
        mode = self.get_netmode_with_retry()

        # 根据模式设置目标路径和描述
        if mode == "故事模式":
            dst_names = ["在线", "进入GTA在线模式", "凭邀请加入的战局"]
            mode_desc = "切换到新的邀请战局（从故事模式）"
        else:  # mode == "在线模式"
            dst_names = ["在线", "寻找新战局", "仅限邀请的战局"]
            mode_desc = "切换到新的邀请战局（从在线模式）"

        # 执行切换
        return self._switch_mode(dst_names, mode_desc, max_retries)

    def sync_to_remote(self):
        """同步存档到远程服务器"""
        logger.note(f"同步存档到远程服务器...")
        self.interactor.close_window()
        time.sleep(10)  # 等待右下角转完圈，出现选项
        self.interactor.cancel(3)
        logger.okay(f"已完成同步，请留意左下角提示: {logstr.mesg('[保存成功]')}")


def parse_args() -> argparse.Namespace:
    """
    解析命令行参数。

    :return: 解析后的参数
    """
    parser = argparse.ArgumentParser(
        description="GTA 在线/故事模式切换工具",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
示例:
  python -m gtaz.workers.mode_switch -g    # 获取当前模式
  python -m gtaz.workers.mode_switch -o    # 故事模式 -> 在线模式
  python -m gtaz.workers.mode_switch -s    # 在线模式 -> 故事模式
  python -m gtaz.workers.mode_switch -i    # 新的邀请战局
  python -m gtaz.workers.mode_switch -y    # 同步存档到远程服务器
        """,
    )

    parser.add_argument(
        "-g",
        "--get",
        action="store_true",
        help="获取当前模式",
    )
    parser.add_argument(
        "-o",
        "--online",
        action="store_true",
        help="故事模式 -> 在线模式",
    )
    parser.add_argument(
        "-s",
        "--story",
        action="store_true",
        help="在线模式 -> 故事模式",
    )
    parser.add_argument(
        "-i",
        "--invite",
        action="store_true",
        help="新的邀请战局",
    )
    parser.add_argument(
        "-y",
        "--sync",
        action="store_true",
        help="同步存档到远程服务器",
    )
    return parser.parse_args()


def main():
    """主函数，处理命令行参数"""
    if len(sys.argv) == 1:
        # 没有参数时显示帮助
        sys.argv.append("-h")
    args = parse_args()

    switcher = NetmodeSwitcher()

    if args.get:
        switcher.get_netmode_with_retry()

    if args.online:
        switcher.switch_story_to_online()

    if args.story:
        switcher.switch_online_to_story()

    if args.invite:
        switcher.switch_to_new_invite_lobby()

    if args.sync:
        switcher.sync_to_remote()


if __name__ == "__main__":
    main()

    # 显示帮助信息
    # python -m gtaz.workers.mode_switch

    # 获取当前模式
    # python -m gtaz.workers.mode_switch -g

    # 故事模式 -> 在线模式
    # python -m gtaz.workers.mode_switch -o

    # 在线模式 -> 故事模式
    # python -m gtaz.workers.mode_switch -s

    # 新的邀请战局
    # python -m gtaz.workers.mode_switch -i

    # 同步存档到远程服务器
    # python -m gtaz.workers.mode_switch -y



"""
GTAV 行为克隆 (Behavior Cloning) 模型训练

基于 ResNet18 的端到端行为克隆模型，用于从小地图帧预测 WASD 按键动作。

架构设计（高内聚、低耦合）:
- Config: 统一配置管理
- FrameSample: 单帧数据结构
- DataProcessor: 数据解析和处理（静态方法）
- DataManager: 数据加载、序列创建和数据集划分
- AgencyMoveDataset: PyTorch 数据集封装
- TemporalResNet/SimpleCNNModel: 模型定义
- ModelFactory: 模型创建工厂
- Metrics: 评估指标计算
- Trainer: 训练流程管理
- Predictor: 推理预测
- AgencyMovePipeline: 顶层流程编排
"""

import argparse
import json
import random
import platform
import re
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from PIL import Image
from tclogger import TCLogger, TCLogbar, logstr, dict_to_str, Runtimer
from torch.utils.data import DataLoader, Dataset
from torchvision import models, transforms


logger = TCLogger(name="AgencyMove", use_prefix=True)


# ===== 常量定义 ===== #

KEY_TO_INDEX = {"W": 0, "A": 1, "S": 2, "D": 3}
INDEX_TO_KEY = {0: "W", 1: "A", 2: "S", 3: "D"}
NUM_KEYS = 4


# ===== 配置类 ===== #

# src/gtaz
SRC_DIR = Path(__file__).parents[2]
DATA_DIR = SRC_DIR / "cache/agency_move"
CKPT_DIR = SRC_DIR / "checkpoints/agency_move"


@dataclass
class Config:
    """统一配置管理"""

    # 数据相关
    data_dir: str = str(DATA_DIR)
    history_frames: int = 4  # 模型输入的历史参考帧数量
    image_size: tuple[int, int] = (160, 220)  # (H, W)
    use_key_history: bool = True
    min_gap_threshold: float = 1.0  # 帧间隔阈值（秒）
    use_image_cache: bool = True  # 是否使用图像缓存（提升性能但增加内存占用）

    # 模型相关
    backbone: str = "resnet18"
    pretrained: bool = True
    hidden_dim: int = 256
    dropout: float = 0.3
    model_type: str = "temporal"  # "temporal" or "simple"

    # 训练相关
    batch_size: int = 32
    num_epochs: int = 50
    learning_rate: Optional[float] = None  # None 表示使用自动调整
    weight_decay: float = 1e-4
    lr_scheduler: str = "cosine"  # "cosine" or "step"
    lr_step_size: int = 10
    lr_gamma: float = 0.5
    class_weights: tuple[float, float, float, float] = (1.0, 2.0, 2.0, 2.0)

    # 数据划分
    train_ratio: float = 0.9
    val_ratio: float = 0.1
    test_ratio: float = 0.1  # 不浪费训练和验证数据，随机选取一定比例测试

    # 其他
    # num_workers: int = 0 if platform.system() == "Windows" else 4
    num_workers: int = 4 if torch.cuda.is_available() else 2  # GPU可用时使用更多worker
    prefetch_factor: int = 4  # 预取因子，每个worker预取的batch数
    persistent_workers: bool = True  # 保持worker进程存活，避免重复创建
    device: str = "cuda" if torch.cuda.is_available() else "cpu"
    seed: int = 42
    save_dir: str = str(CKPT_DIR)
    log_interval: int = 10
    threshold: float = 0.5  # 预测阈值
    overwrite: bool = False  # 是否覆盖旧模型，不使用checkpoint恢复

    def to_dict(self) -> dict:
        """转换为字典"""
        return {k: v for k, v in self.__dict__.items()}

    @classmethod
    def from_dict(cls, d: dict) -> "Config":
        """从字典创建"""
        return cls(**{k: v for k, v in d.items() if k in cls.__dataclass_fields__})

    def get_model_name(self) -> str:
        """生成模型名称（基于训练参数）"""
        # 构建参数字符串
        params = [
            f"{self.model_type}",
            f"f{self.history_frames}",
            f"b{self.batch_size}",
            f"e{self.num_epochs}",
        ]

        # 添加学习率信息
        if self.learning_rate is None:
            params.append("lr_auto")
        else:
            lr_str = f"{self.learning_rate:.0e}".replace("e-0", "e-").replace(
                "e+0", "e"
            )
            params.append(f"lr{lr_str}")

        # 添加dropout信息（如果不是默认值）
        if self.dropout != 0.3:
            params.append(f"drop{self.dropout:.2f}".replace("0.", "."))

        # 添加hidden_dim信息（如果不是默认值）
        if self.hidden_dim != 256:
            params.append(f"h{self.hidden_dim}")

        return f"agency_move_{'_'.join(params)}"


# ===== 数据结构 ===== #


@dataclass
class FrameSample:
    """单帧数据样本"""

    image_path: str
    json_path: str
    frame_index: int
    timestamp: float
    keys_pressed: list[str] = field(default_factory=list)
    key_vector: np.ndarray = field(default_factory=lambda: np.zeros(NUM_KEYS))


# ===== 数据处理器（静态方法集合） ===== #


class DataProcessor:
    """数据解析和处理器（静态工具类）"""

    @staticmethod
    def parse_filename(filename: str) -> tuple[str, int]:
        """解析文件名，提取时间戳和帧索引"""
        pattern = r"(\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}-\d{3})_(\d+)"
        match = re.match(pattern, filename)
        if match:
            return match.group(1), int(match.group(2))
        return "", 0

    @staticmethod
    def is_segment_start(filename: str) -> bool:
        """判断文件名是否标志着新段的开始（末尾为 _0001 或 _001 等）"""
        pattern = r"_0+1\.(json|jpg)$"
        return bool(re.search(pattern, filename))

    @staticmethod
    def load_json(json_path: str) -> dict:
        """加载 JSON 数据文件"""
        with open(json_path, "r", encoding="utf-8") as f:
            return json.load(f)

    @staticmethod
    def extract_key_vector(json_data: dict) -> np.ndarray:
        """从 JSON 数据中提取按键向量"""
        key_vector = np.zeros(NUM_KEYS, dtype=np.float32)
        if json_data.get("has_action", False) and "keys" in json_data:
            for key_info in json_data["keys"]:
                key_name = key_info.get("key_name", "")
                is_pressed = key_info.get("is_pressed", False)
                if key_name in KEY_TO_INDEX and is_pressed:
                    key_vector[KEY_TO_INDEX[key_name]] = 1.0
        return key_vector

    @classmethod
    def parse_frame(cls, json_path: Path) -> Optional[FrameSample]:
        """解析单帧数据"""
        image_path = json_path.with_suffix(".jpg")
        if not image_path.exists():
            return None

        try:
            json_data = cls.load_json(str(json_path))
        except (json.JSONDecodeError, IOError):
            return None

        _, frame_index = cls.parse_filename(json_path.name)
        timestamp = json_data.get("time", {}).get("timestamp", 0)
        key_vector = cls.extract_key_vector(json_data)
        keys_pressed = [INDEX_TO_KEY[i] for i in range(NUM_KEYS) if key_vector[i] > 0]

        return FrameSample(
            image_path=str(image_path),
            json_path=str(json_path),
            frame_index=frame_index,
            timestamp=timestamp,
            keys_pressed=keys_pressed,
            key_vector=key_vector,
        )

    @classmethod
    def collect_session_frames(cls, session_dir: Path) -> list[list[FrameSample]]:
        """收集一个会话目录中的所有帧数据，按段分组"""
        json_files = sorted(session_dir.glob("*.json"))

        segments = []
        current_segment = []

        for json_path in json_files:
            # 如果是新段的开始且当前段不为空，保存当前段
            if cls.is_segment_start(json_path.name) and current_segment:
                segments.append(current_segment)
                current_segment = []

            sample = cls.parse_frame(json_path)
            if sample:
                current_segment.append(sample)

        # 保存最后一段
        if current_segment:
            segments.append(current_segment)

        # 按时间戳排序每个段内的样本
        for segment in segments:
            segment.sort(key=lambda x: x.timestamp)

        return segments


# ===== 数据集类 ===== #


class AgencyMoveDataset(Dataset):
    """GTAV 行为克隆数据集"""

    def __init__(
        self,
        sequences: list[list[FrameSample]],
        history_frames: int = 4,
        use_key_history: bool = True,
        image_size: tuple[int, int] = (160, 220),
        augment: bool = False,
        use_cache: bool = True,
    ):
        self.sequences = sequences
        self.history_frames = history_frames
        self.use_key_history = use_key_history
        self.image_size = image_size
        self.augment = augment
        self.use_cache = use_cache
        self.transform = self._build_transform()
        # 缓存预处理后的tensor而非PIL Image，减少CPU处理
        self._tensor_cache = {} if use_cache else None

    def _build_transform(self) -> transforms.Compose:
        """构建图像变换"""
        if self.augment:
            return transforms.Compose(
                [
                    transforms.Resize(self.image_size),
                    transforms.ColorJitter(
                        brightness=0.1, contrast=0.1, saturation=0.1
                    ),
                    transforms.RandomAffine(degrees=0, translate=(0.02, 0.02)),
                    transforms.ToTensor(),
                    transforms.Normalize(
                        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                    ),
                ]
            )
        return transforms.Compose(
            [
                transforms.Resize(self.image_size),
                transforms.ToTensor(),
                transforms.Normalize(
                    mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
                ),
            ]
        )

    def __len__(self) -> int:
        return len(self.sequences)

    def warmup_cache(self):
        """预热缓存：预先加载所有图像到缓存中（仅对验证/测试集）"""
        if not self.use_cache or self.augment:
            return

        # 收集所有唯一的图像路径
        unique_paths = set()
        for seq in self.sequences:
            for sample in seq:
                unique_paths.add(sample.image_path)

        if not unique_paths:
            return

        logger.note(f"预热缓存: 加载 {len(unique_paths)} 张图像...")
        warmup_desc = logstr.note("* 预热缓存")
        bar = TCLogbar(total=len(unique_paths), desc=warmup_desc)

        for img_path in unique_paths:
            if img_path not in self._tensor_cache:
                img = Image.open(img_path).convert("RGB")
                img_tensor = self.transform(img)
                self._tensor_cache[img_path] = img_tensor
            bar.update(1)

        bar.update(flush=True, linebreak=True)
        logger.okay(f"缓存预热完成: {len(self._tensor_cache)} 张图像")

    def _load_and_transform_image(self, image_path: str) -> torch.Tensor:
        """加载并转换图像为tensor，支持缓存"""
        # 对于验证/测试集，缓存预处理后的tensor
        if self.use_cache and not self.augment and image_path in self._tensor_cache:
            return self._tensor_cache[image_path]

        # 加载并转换图像
        img = Image.open(image_path).convert("RGB")
        img_tensor = self.transform(img)

        # 对于验证/测试集缓存tensor（训练集有数据增强所以不缓存）
        if self.use_cache and not self.augment:
            self._tensor_cache[image_path] = img_tensor

        return img_tensor

    def __getitem__(self, idx: int) -> dict:
        sequence = self.sequences[idx]

        # 加载并处理图像 - 直接获取tensor
        images = [self._load_and_transform_image(s.image_path) for s in sequence]
        images = torch.stack(images, dim=0)

        # 获取按键信息 - 预分配numpy数组提升效率
        seq_len = len(sequence)
        key_vectors = np.empty((seq_len, NUM_KEYS), dtype=np.float32)
        for i, s in enumerate(sequence):
            key_vectors[i] = s.key_vector

        key_history = torch.from_numpy(key_vectors)
        target = key_history[-1]

        result = {"images": images, "target": target}
        if self.use_key_history:
            result["key_history"] = key_history[:-1]

        return result


# ===== 数据管理器 ===== #


class DataManager:
    """数据管理器：负责数据加载、序列创建和数据集划分"""

    def __init__(self, config: Config):
        self.config = config
        self._sequences: list[list[FrameSample]] = []

    @property
    def sequences(self) -> list[list[FrameSample]]:
        return self._sequences

    def create_sequences(self, samples: list[FrameSample]) -> list[list[FrameSample]]:
        """将帧样本按历史帧数量分割成连续的序列"""
        history_frames = self.config.history_frames
        min_gap = self.config.min_gap_threshold

        if len(samples) < history_frames:
            return []

        sequences = []
        current_seq = [samples[0]]

        for i in range(1, len(samples)):
            time_gap = samples[i].timestamp - samples[i - 1].timestamp
            if time_gap > min_gap or time_gap < 0:
                current_seq = [samples[i]]
            else:
                current_seq.append(samples[i])

            if len(current_seq) >= history_frames:
                for j in range(len(current_seq) - history_frames + 1):
                    sequences.append(current_seq[j : j + history_frames])
                current_seq = current_seq[-(history_frames - 1) :]

        return sequences

    def load_all_data(self) -> list[list[FrameSample]]:
        """加载所有训练数据"""
        data_dir = Path(self.config.data_dir)
        all_sequences = []

        session_dirs = [d for d in data_dir.iterdir() if d.is_dir()]
        logger.note(f"找到 {logstr.mesg(len(session_dirs))} 个会话目录")

        total_segments = 0

        load_desc = logstr.note("* 加载数据")
        bar = TCLogbar(total=len(session_dirs), desc=load_desc)
        for session_dir in session_dirs:
            segments = DataProcessor.collect_session_frames(session_dir)
            total_segments += len(segments)

            # 对每个段分别创建序列
            for segment in segments:
                if len(segment) >= self.config.history_frames:
                    sequences = self.create_sequences(segment)
                    all_sequences.extend(sequences)
            bar.update(1)
        bar.update(flush=True, linebreak=True)

        self._sequences = all_sequences
        logger.okay(
            f"从 {logstr.mesg(total_segments)} 个数据段中"
            f"总共生成 {logstr.mesg(len(all_sequences))} 个训练样本"
        )
        return all_sequences

    def split_data(self) -> tuple[list, list, list]:
        """划分数据集"""
        if not self._sequences:
            raise ValueError("没有可用的序列数据！请先调用 load_all_data()")

        sequences = self._sequences.copy()
        random.shuffle(sequences)

        total = len(sequences)
        train_size = int(total * self.config.train_ratio)

        train_seqs = sequences[:train_size]
        val_seqs = sequences[train_size:]

        test_size = int(total * self.config.test_ratio)
        test_seqs = sequences[-test_size:]

        logger.note("> 数据集划分:")
        split_info = {
            "训练集": len(train_seqs),
            "验证集": len(val_seqs),
            "测试集": len(test_seqs),
        }
        logger.mesg(dict_to_str(split_info), indent=2)
        return train_seqs, val_seqs, test_seqs

    def create_datasets(
        self,
    ) -> tuple[AgencyMoveDataset, AgencyMoveDataset, AgencyMoveDataset]:
        """创建数据集"""
        train_seqs, val_seqs, test_seqs = self.split_data()

        train_ds = AgencyMoveDataset(
            train_seqs,
            self.config.history_frames,
            self.config.use_key_history,
            self.config.image_size,
            augment=True,
            use_cache=self.config.use_image_cache,
        )
        val_ds = AgencyMoveDataset(
            val_seqs,
            self.config.history_frames,
            self.config.use_key_history,
            self.config.image_size,
            augment=False,
            use_cache=self.config.use_image_cache,
        )
        test_ds = AgencyMoveDataset(
            test_seqs,
            self.config.history_frames,
            self.config.use_key_history,
            self.config.image_size,
            augment=False,
            use_cache=self.config.use_image_cache,
        )

        # 预热验证和测试集缓存
        if self.config.use_image_cache:
            val_ds.warmup_cache()
            test_ds.warmup_cache()

        return train_ds, val_ds, test_ds

    def create_data_loaders(self) -> tuple[DataLoader, DataLoader, DataLoader]:
        """创建数据加载器"""
        if not self._sequences:
            self.load_all_data()

        train_ds, val_ds, test_ds = self.create_datasets()

        # 优化的DataLoader配置
        common_kwargs = {
            "num_workers": self.config.num_workers,
            "pin_memory": True,
            "prefetch_factor": (
                self.config.prefetch_factor if self.config.num_workers > 0 else None
            ),
            "persistent_workers": (
                self.config.persistent_workers if self.config.num_workers > 0 else False
            ),
        }

        train_loader = DataLoader(
            train_ds,
            batch_size=self.config.batch_size,
            shuffle=True,
            drop_last=True,
            **common_kwargs,
        )
        val_loader = DataLoader(
            val_ds,
            batch_size=self.config.batch_size,
            shuffle=False,
            **common_kwargs,
        )
        test_loader = DataLoader(
            test_ds,
            batch_size=self.config.batch_size,
            shuffle=False,
            **common_kwargs,
        )

        return train_loader, val_loader, test_loader

    def analyze(self) -> dict:
        """分析数据集统计信息"""
        if not self._sequences:
            self.load_all_data()

        key_counts = {key: 0 for key in INDEX_TO_KEY.values()}
        key_combo_counts = {}
        total = len(self._sequences)

        for seq in self._sequences:
            sample = seq[-1]
            keys = tuple(sorted(sample.keys_pressed))
            key_combo_counts[keys] = key_combo_counts.get(keys, 0) + 1
            for key in sample.keys_pressed:
                key_counts[key] += 1

        # 打印统计信息
        logger.note(f"> 数据集统计: 总样本数 {logstr.mesg(total)}")

        logger.note("单键分布:")
        key_dist = {}
        for key, count in key_counts.items():
            ratio = f"{count / total * 100:.2f}%" if total > 0 else "0%"
            key_dist[key] = f"{count} ({ratio})"
        logger.mesg(dict_to_str(key_dist), indent=2)

        logger.note("按键组合分布 (Top 10):")
        combo_dist = {}
        for combo, count in sorted(key_combo_counts.items(), key=lambda x: -x[1])[:10]:
            combo_str = "+".join(combo) if combo else "(无按键)"
            ratio = f"{count / total * 100:.2f}%" if total > 0 else "0%"
            combo_dist[combo_str] = f"{count} ({ratio})"
        logger.mesg(dict_to_str(combo_dist), indent=2)

        return {
            "total_samples": total,
            "key_counts": key_counts,
            "key_combo_counts": key_combo_counts,
        }


# ===== 模型定义 ===== #


class TemporalResNet(nn.Module):
    """时序 ResNet 模型：ResNet18 + GRU + 历史按键嵌入"""

    def __init__(
        self,
        history_frames: int = 4,
        num_keys: int = NUM_KEYS,
        hidden_dim: int = 256,
        use_key_history: bool = True,
        dropout: float = 0.3,
        pretrained: bool = True,
    ):
        super().__init__()
        self.history_frames = history_frames
        self.num_keys = num_keys
        self.use_key_history = use_key_history
        self.hidden_dim = hidden_dim
        self.feature_dim = 512

        # ResNet18 backbone
        resnet = models.resnet18(
            weights=models.ResNet18_Weights.DEFAULT if pretrained else None
        )
        self.backbone = nn.Sequential(*list(resnet.children())[:-2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        # 时序融合层
        self.temporal_gru = nn.GRU(
            input_size=self.feature_dim,
            hidden_size=hidden_dim,
            num_layers=1,
            batch_first=True,
            bidirectional=False,
        )

        # 历史按键嵌入
        fc_input_dim = hidden_dim
        if use_key_history:
            self.key_embedding = nn.Sequential(
                nn.Linear((history_frames - 1) * num_keys, 64),
                nn.ReLU(),
                nn.Dropout(dropout),
            )
            fc_input_dim += 64
        else:
            self.key_embedding = None

        # 输出头
        self.fc = nn.Sequential(
            nn.Linear(fc_input_dim, hidden_dim),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, num_keys),
        )

    def forward(
        self, images: torch.Tensor, key_history: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        batch_size, num_frames_in_batch, C, H, W = images.shape

        # 提取特征
        images_flat = images.view(batch_size * num_frames_in_batch, C, H, W)
        features = self.backbone(images_flat)
        features = self.avgpool(features).view(batch_size * num_frames_in_batch, -1)
        features = features.view(batch_size, num_frames_in_batch, self.feature_dim)

        # GRU 时序建模
        _, hidden = self.temporal_gru(features)
        temporal_features = hidden.squeeze(0)

        # 融合按键历史
        if self.use_key_history and key_history is not None:
            key_flat = key_history.view(batch_size, -1)
            key_features = self.key_embedding(key_flat)
            combined = torch.cat([temporal_features, key_features], dim=1)
        else:
            combined = temporal_features

        return self.fc(combined)


class SimpleCNNModel(nn.Module):
    """简化 CNN 模型：将多个历史帧图像在通道维度堆叠"""

    def __init__(
        self,
        history_frames: int = 4,
        num_keys: int = NUM_KEYS,
        hidden_dim: int = 256,
        use_key_history: bool = True,
        dropout: float = 0.3,
        pretrained: bool = True,
    ):
        super().__init__()
        self.history_frames = history_frames
        self.num_keys = num_keys
        self.use_key_history = use_key_history

        # 修改 ResNet18 第一层
        resnet = models.resnet18(
            weights=models.ResNet18_Weights.DEFAULT if pretrained else None
        )
        original_conv = resnet.conv1
        self.conv1 = nn.Conv2d(
            history_frames * 3, 64, kernel_size=7, stride=2, padding=3, bias=False
        )

        if pretrained:
            with torch.no_grad():
                weight = (
                    original_conv.weight.data.repeat(1, history_frames, 1, 1)
                    / history_frames
                )
                self.conv1.weight.data = weight

        self.backbone = nn.Sequential(
            resnet.bn1,
            resnet.relu,
            resnet.maxpool,
            resnet.layer1,
            resnet.layer2,
            resnet.layer3,
            resnet.layer4,
            resnet.avgpool,
        )

        fc_input_dim = 512
        if use_key_history:
            self.key_embedding = nn.Sequential(
                nn.Linear((history_frames - 1) * num_keys, 64),
                nn.ReLU(),
                nn.Dropout(dropout),
            )
            fc_input_dim += 64
        else:
            self.key_embedding = None

        self.fc = nn.Sequential(
            nn.Linear(fc_input_dim, hidden_dim),
            nn.ReLU(),
            nn.Dropout(dropout),
            nn.Linear(hidden_dim, num_keys),
        )

    def forward(
        self, images: torch.Tensor, key_history: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        batch_size, num_frames_in_batch, C, H, W = images.shape
        images = images.view(batch_size, num_frames_in_batch * C, H, W)

        x = self.conv1(images)
        x = self.backbone(x)
        features = x.view(batch_size, -1)

        if self.use_key_history and key_history is not None:
            key_flat = key_history.view(batch_size, -1)
            key_features = self.key_embedding(key_flat)
            combined = torch.cat([features, key_features], dim=1)
        else:
            combined = features

        return self.fc(combined)


# ===== 模型工厂 ===== #


class ModelFactory:
    """模型创建工厂"""

    @staticmethod
    def create(config: Config) -> nn.Module:
        """根据配置创建模型"""
        model_cls = (
            TemporalResNet if config.model_type == "temporal" else SimpleCNNModel
        )
        return model_cls(
            history_frames=config.history_frames,
            num_keys=NUM_KEYS,
            hidden_dim=config.hidden_dim,
            use_key_history=config.use_key_history,
            dropout=config.dropout,
            pretrained=config.pretrained,
        )

    @staticmethod
    def load(model_path: str, device: torch.device = None) -> tuple[nn.Module, dict]:
        """加载模型"""
        if device is None:
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        checkpoint = torch.load(model_path, map_location=device, weights_only=False)
        config_dict = checkpoint.get("config", {})

        model = TemporalResNet(
            history_frames=config_dict.get("history_frames", 4),
            num_keys=NUM_KEYS,
            hidden_dim=config_dict.get("hidden_dim", 256),
            use_key_history=config_dict.get("use_key_history", True),
            dropout=config_dict.get("dropout", 0.3),
            pretrained=False,
        )
        model.load_state_dict(checkpoint["model_state_dict"])
        model = model.to(device)
        model.eval()

        return model, config_dict


# ===== 评估指标 ===== #


class Metrics:
    """评估指标计算器"""

    def __init__(self, num_keys: int = NUM_KEYS, threshold: float = 0.5):
        self.num_keys = num_keys
        self.threshold = threshold
        self.reset()

    def reset(self):
        self.tp = np.zeros(self.num_keys)
        self.fp = np.zeros(self.num_keys)
        self.tn = np.zeros(self.num_keys)
        self.fn = np.zeros(self.num_keys)
        self.exact_match = 0
        self.total = 0

    def update(self, predictions: torch.Tensor, targets: torch.Tensor):
        probs = torch.sigmoid(predictions).detach().cpu().numpy()
        preds = (probs > self.threshold).astype(np.float32)
        targets_np = targets.detach().cpu().numpy()

        for i in range(self.num_keys):
            p, t = preds[:, i], targets_np[:, i]
            self.tp[i] += np.sum((p == 1) & (t == 1))
            self.fp[i] += np.sum((p == 1) & (t == 0))
            self.tn[i] += np.sum((p == 0) & (t == 0))
            self.fn[i] += np.sum((p == 0) & (t == 1))

        self.exact_match += np.sum(np.all(preds == targets_np, axis=1))
        self.total += len(targets_np)

    def compute(self) -> dict:
        results = {}
        for i, key_name in INDEX_TO_KEY.items():
            precision = self.tp[i] / (self.tp[i] + self.fp[i] + 1e-8)
            recall = self.tp[i] / (self.tp[i] + self.fn[i] + 1e-8)
            f1 = 2 * precision * recall / (precision + recall + 1e-8)
            accuracy = (self.tp[i] + self.tn[i]) / (
                self.tp[i] + self.tn[i] + self.fp[i] + self.fn[i] + 1e-8
            )
            results.update(
                {
                    f"{key_name}_precision": precision,
                    f"{key_name}_recall": recall,
                    f"{key_name}_f1": f1,
                    f"{key_name}_accuracy": accuracy,
                }
            )

        results["avg_precision"] = np.mean(
            [results[f"{k}_precision"] for k in INDEX_TO_KEY.values()]
        )
        results["avg_recall"] = np.mean(
            [results[f"{k}_recall"] for k in INDEX_TO_KEY.values()]
        )
        results["avg_f1"] = np.mean([results[f"{k}_f1"] for k in INDEX_TO_KEY.values()])
        results["avg_accuracy"] = np.mean(
            [results[f"{k}_accuracy"] for k in INDEX_TO_KEY.values()]
        )
        results["exact_match_accuracy"] = self.exact_match / (self.total + 1e-8)

        return results


# ===== 训练器 ===== #


class Trainer:
    """训练器：管理模型训练流程"""

    def __init__(self, config: Config, model: nn.Module, device: torch.device = None):
        self.config = config
        self.model = model
        self.device = device or torch.device(config.device)
        self.model.to(self.device)

        # 混合精度训练（仅GPU可用）
        self.use_amp = torch.cuda.is_available()
        self.scaler = torch.amp.GradScaler("cuda") if self.use_amp else None

        # 编译模型以提升GPU利用率（PyTorch 2.0+，仅Linux）
        # Windows上torch.compile需要Triton但不可用，因此禁用
        self.model_compiled = False
        if (
            torch.cuda.is_available()
            and hasattr(torch, "compile")
            and platform.system() != "Windows"
        ):
            try:
                self.model = torch.compile(self.model, mode="default")
                self.model_compiled = True
                logger.note("模型编译成功")
            except Exception as e:
                logger.warn(f"模型编译失败，使用普通模式: {e}")
                self.model_compiled = False

        # 损失函数
        self.class_weights = torch.tensor(config.class_weights, dtype=torch.float32).to(
            self.device
        )
        self.criterion = nn.BCEWithLogitsLoss(pos_weight=self.class_weights)

        # 优化器和学习率调度器
        if config.learning_rate is None:
            # 使用自动调整学习率：从较大的初始值开始
            initial_lr = 1e-4
            self.optimizer = torch.optim.AdamW(
                model.parameters(),
                lr=initial_lr,
                weight_decay=config.weight_decay,
            )
            # 使用 ReduceLROnPlateau：根据验证指标自动调整
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(
                self.optimizer,
                mode="max",  # 监控 F1 分数（越高越好）
                factor=0.5,  # 学习率减半
                patience=5,  # 5 个 epoch 没有提升则降低
                min_lr=1e-6,  # 最小学习率
            )
            self.use_plateau_scheduler = True
            logger.note(f"使用自动学习率调整，初始学习率: {logstr.mesg(initial_lr)}")
        else:
            # 使用固定学习率
            self.optimizer = torch.optim.AdamW(
                model.parameters(),
                lr=config.learning_rate,
                weight_decay=config.weight_decay,
            )
            # 学习率调度器
            if config.lr_scheduler == "cosine":
                self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                    self.optimizer, T_max=config.num_epochs
                )
            else:
                self.scheduler = torch.optim.lr_scheduler.StepLR(
                    self.optimizer, step_size=config.lr_step_size, gamma=config.lr_gamma
                )
            self.use_plateau_scheduler = False

        self.history = {
            "train_loss": [],
            "val_loss": [],
            "train_f1": [],
            "val_f1": [],
            "train_exact_match": [],
            "val_exact_match": [],
        }
        self.latest_val_f1 = 0.0
        self.latest_epoch = 0
        self.start_epoch = 0
        self.best_val_f1 = 0.0  # 跟踪历史最佳验证F1

        # 跟踪最佳指标（用于计算百分比变化）
        self.best_metrics = {
            "train_loss": float("inf"),
            "val_loss": float("inf"),
            "train_f1": 0.0,
            "val_f1": 0.0,
            "train_exact_match": 0.0,
            "val_exact_match": 0.0,
        }

    def _format_metric_change(
        self, current: float, best: float, is_loss: bool = False
    ) -> str:
        """格式化指标变化（百分比，相对于最佳值）

        Args:
            current: 当前值
            best: 最佳值
            is_loss: 是否为损失指标（True表示越小越好，False表示越大越好）
        """
        if best == 0 or (is_loss and best == float("inf")):
            return ""

        percent_diff = ((current - best) / abs(best)) * 100
        abs_percent_diff = abs(percent_diff)

        # 对于loss，负数表示改善；对于其他指标，正数表示改善
        is_better = (percent_diff < 0) if is_loss else (percent_diff > 0)

        if abs_percent_diff < 0.05:
            return f" ({logstr.mesg(f'≈ best')})"
        elif is_better:
            return f" ({logstr.okay(f'↑ {abs_percent_diff:.1f}%')})"
        else:
            return f" ({logstr.warn(f'↓ {abs_percent_diff:.1f}%')})"

    def _log_checkpoint_info(self, checkpoint_path: str):
        """记录checkpoint恢复信息"""
        if self.latest_epoch >= self.config.num_epochs:
            logger.okay(f"> 训练已完成")
        else:
            logger.note(f"> 从 checkpoint 恢复训练:")
        info_dict = {
            "路径": checkpoint_path,
            "起始轮次": self.start_epoch,
            "上次F1": f"{self.latest_val_f1:.4f} (Epoch {self.latest_epoch})",
            "最佳F1": f"{self.best_val_f1:.4f}",
            "学习率": f"{self.optimizer.param_groups[0]['lr']:.6f}",
            "混合精度": "开启" if self.use_amp else "关闭",
            "模型编译": "开启" if self.model_compiled else "关闭",
        }
        logger.mesg(dict_to_str(info_dict), indent=2)

    def _log_epoch_metrics(
        self,
        epoch: int,
        train_loss: float,
        train_metrics: dict,
        val_loss: float,
        val_metrics: dict,
    ):
        """记录epoch训练和验证指标"""
        current_lr = self.optimizer.param_groups[0]["lr"]
        logger.mesg(f"学习率: {logstr.file(f'{current_lr:.6f}')}")

        # 计算训练指标的变化（相对于最佳值）
        train_loss_change = ""
        train_f1_change = ""
        train_exact_change = ""
        if epoch > 0:
            train_loss_change = self._format_metric_change(
                train_loss, self.best_metrics["train_loss"], is_loss=True
            )
            train_f1_change = self._format_metric_change(
                train_metrics["avg_f1"], self.best_metrics["train_f1"], is_loss=False
            )
            train_exact_change = self._format_metric_change(
                train_metrics["exact_match_accuracy"],
                self.best_metrics["train_exact_match"],
                is_loss=False,
            )

        # 计算验证指标的变化（相对于最佳值）
        val_loss_change = ""
        val_f1_change = ""
        val_exact_change = ""
        if epoch > 0:
            val_loss_change = self._format_metric_change(
                val_loss, self.best_metrics["val_loss"], is_loss=True
            )
            val_f1_change = self._format_metric_change(
                val_metrics["avg_f1"], self.best_metrics["val_f1"], is_loss=False
            )
            val_exact_change = self._format_metric_change(
                val_metrics["exact_match_accuracy"],
                self.best_metrics["val_exact_match"],
                is_loss=False,
            )

        # 打印训练指标
        logger.mesg(
            f"训练 - Loss: {logstr.file(f'{train_loss:.4f}')}{train_loss_change}, "
            f"Avg F1: {logstr.file(f'{train_metrics['avg_f1']:.4f}')}{train_f1_change}, "
            f"Exact Match: {logstr.file(f'{train_metrics['exact_match_accuracy']:.4f}')}{train_exact_change}"
        )

        # 打印验证指标
        logger.mesg(
            f"验证 - Loss: {logstr.file(f'{val_loss:.4f}')}{val_loss_change}, "
            f"Avg F1: {logstr.file(f'{val_metrics['avg_f1']:.4f}')}{val_f1_change}, "
            f"Exact Match: {logstr.file(f'{val_metrics['exact_match_accuracy']:.4f}')}{val_exact_change}"
        )

    def _format_key_f1_scores(self, metrics: dict, prefix: str = "") -> str:
        """格式化各按键的F1分数为单行字符串

        Args:
            metrics: 包含各键F1分数的指标字典
            prefix: 前缀文本（如"训练"、"验证"、"测试"）

        Returns:
            格式化的字符串，如 "训练: W: 0.9886, A: 0.8992, S: 0.7273, D: 0.8068"
        """
        key_scores = []
        for key_name in INDEX_TO_KEY.values():
            f1_score = metrics[f"{key_name}_f1"]
            key_str = logstr.note(key_name)
            score_str = logstr.hint(f"{f1_score:.4f}")
            key_scores.append(f"{key_str}: {score_str}")
        scores_str = ", ".join(key_scores)
        if prefix:
            return f"{prefix}: {scores_str}"
        return scores_str

    def _log_per_key_metrics(self, train_metrics: dict, val_metrics: dict):
        """记录各按键的F1分数"""
        train_line = self._format_key_f1_scores(train_metrics, "训练")
        val_line = self._format_key_f1_scores(val_metrics, "验证")
        logger.mesg(train_line)
        logger.mesg(val_line)

    def _log_model_save(self, save_path: Path):
        """记录模型保存信息"""
        logger.okay(f"保存最新模型:")
        logger.file(f"{save_path}")

    def _save_latest_checkpoint(
        self, save_dir: Path, epoch: int, val_f1: float, val_loss: float
    ):
        """保存最新模型检查点"""
        model_name = self.config.get_model_name()
        save_path = save_dir / f"{model_name}.pth"
        self._save_checkpoint(save_path, epoch, val_f1, val_loss, is_best=False)
        self._log_model_save(save_path)

    def _save_best_checkpoint(
        self, save_dir: Path, epoch: int, val_f1: float, val_loss: float
    ):
        """保存最佳模型检查点（仅当F1提升时）"""
        model_name = self.config.get_model_name()
        best_save_path = save_dir / f"{model_name}_best.pth"
        self._save_checkpoint(best_save_path, epoch, val_f1, val_loss, is_best=True)
        logger.okay(f"保存最佳模型 (F1: {logstr.mesg(f'{self.best_val_f1:.4f}')})")
        logger.file(f"{best_save_path}")

    def _log_training_config(self):
        """记录训练配置信息"""
        logger.note("> 训练配置:")
        params_count = sum(p.numel() for p in self.model.parameters())
        params_count_str = f"{params_count/1e6:.1f}M"
        train_info = {
            "设备": str(self.device),
            "参数量": params_count_str,
            "混合精度": "开启" if self.use_amp else "关闭",
            "模型编译": "开启" if self.model_compiled else "关闭",
            "批次大小": self.config.batch_size,
            "工作进程": self.config.num_workers,
            "预取因子": self.config.prefetch_factor,
        }
        if self.start_epoch > 0 and self.start_epoch < self.config.num_epochs:
            train_info["状态"] = logstr.mesg(
                f"继续训练 (从 Epoch {self.start_epoch + 1} 开始)"
            )
        elif self.start_epoch >= self.config.num_epochs:
            train_info["状态"] = logstr.okay(f"训练已完成")
        else:
            train_info["状态"] = logstr.mesg(f"从头开始训练")
        logger.mesg(dict_to_str(train_info), indent=2)
        print()

    def _log_training_summary(self):
        """记录训练总结信息"""
        # 找到验证集最佳epoch
        best_val_f1_idx = self.history["val_f1"].index(max(self.history["val_f1"]))
        best_epoch = best_val_f1_idx + 1
        best_val_f1 = self.history["val_f1"][best_val_f1_idx]
        best_val_loss = self.history["val_loss"][best_val_f1_idx]
        best_val_exact = self.history["val_exact_match"][best_val_f1_idx]

        logger.okay("> 训练完成！")
        sep_str = f"\n{'=' * 70}\n"
        logger.okay(
            f"{sep_str}"
            f"最佳验证结果 (Epoch {logstr.mesg(best_epoch)}): "
            f"Loss={logstr.file(f'{best_val_loss:.4f}')}, "
            f"F1={logstr.mesg(f'{best_val_f1:.4f}')}, "
            f"Exact Match={logstr.file(f'{best_val_exact:.4f}')}"
            f"{sep_str}",
        )

    def _save_training_history(self, save_dir: Path):
        """保存训练历史到文件"""
        model_name = self.config.get_model_name()
        history_path = save_dir / f"{model_name}.history.json"
        with open(history_path, "w") as f:
            json.dump(self.history, f, indent=2)
        logger.okay(f"保存训练历史:")
        logger.file(f"{history_path}")

    def load_checkpoint(self, checkpoint_path: str) -> bool:
        """从checkpoint恢复训练状态"""
        try:
            checkpoint = torch.load(
                checkpoint_path, map_location=self.device, weights_only=False
            )
            self.model.load_state_dict(checkpoint["model_state_dict"])
            self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])

            # 恢复调度器状态（如果存在）
            if "scheduler_state_dict" in checkpoint and self.scheduler is not None:
                self.scheduler.load_state_dict(checkpoint["scheduler_state_dict"])

            # 恢复训练历史
            if "history" in checkpoint:
                self.history = checkpoint["history"]

                # 从历史记录中恢复最佳指标
                for key in ["train_loss", "val_loss"]:
                    if self.history[key]:
                        self.best_metrics[key] = min(self.history[key])

                for key in [
                    "train_f1",
                    "val_f1",
                    "train_exact_match",
                    "val_exact_match",
                ]:
                    if self.history[key]:
                        self.best_metrics[key] = max(self.history[key])

            # 恢复最新指标
            self.latest_val_f1 = checkpoint.get(
                "latest_val_f1", checkpoint.get("val_f1", 0.0)
            )
            self.latest_epoch = checkpoint.get(
                "latest_epoch", checkpoint.get("epoch", 0)
            )
            self.start_epoch = checkpoint.get("epoch", 0)
            self.best_val_f1 = checkpoint.get("best_val_f1", 0.0)

            self._log_checkpoint_info(checkpoint_path)
            return True
        except Exception as e:
            logger.fail(f"加载 checkpoint 失败: {e}")
            return False

    def train_epoch(self, train_loader: DataLoader) -> tuple[float, dict]:
        """训练一个 epoch"""
        self.model.train()
        total_loss = 0.0
        metrics = Metrics()
        num_batches = len(train_loader)

        train_desc = logstr.note("* Training  ")
        bar = TCLogbar(total=num_batches, desc=train_desc)

        for batch_idx, batch in enumerate(train_loader):
            # 使用非阻塞数据传输
            images = batch["images"].to(self.device, non_blocking=True)
            targets = batch["target"].to(self.device, non_blocking=True)
            key_history = batch.get("key_history")
            if key_history is not None:
                key_history = key_history.to(self.device, non_blocking=True)

            self.optimizer.zero_grad(set_to_none=True)

            # 混合精度训练
            if self.use_amp:
                with torch.amp.autocast("cuda"):
                    logits = self.model(images, key_history)
                    loss = self.criterion(logits, targets)

                self.scaler.scale(loss).backward()
                self.scaler.step(self.optimizer)
                self.scaler.update()
            else:
                logits = self.model(images, key_history)
                loss = self.criterion(logits, targets)
                loss.backward()
                self.optimizer.step()

            # 减少.item()调用频率，累积loss
            total_loss += loss.detach()
            metrics.update(logits.detach(), targets)
            bar.update(1)

        bar.update(flush=True, linebreak=True)

        return (total_loss / num_batches).item(), metrics.compute()

    def validate(self, val_loader: DataLoader) -> tuple[float, dict]:
        """验证模型"""
        self.model.eval()
        total_loss = 0.0
        metrics = Metrics()

        val_desc = logstr.note("* Validation")
        bar = TCLogbar(total=len(val_loader), desc=val_desc)
        with torch.no_grad():
            for batch in val_loader:
                # 使用非阻塞数据传输
                images = batch["images"].to(self.device, non_blocking=True)
                targets = batch["target"].to(self.device, non_blocking=True)
                key_history = batch.get("key_history")
                if key_history is not None:
                    key_history = key_history.to(self.device, non_blocking=True)

                # 混合精度推理
                if self.use_amp:
                    with torch.amp.autocast("cuda"):
                        logits = self.model(images, key_history)
                        loss = self.criterion(logits, targets)
                else:
                    logits = self.model(images, key_history)
                    loss = self.criterion(logits, targets)

                total_loss += loss.item()
                metrics.update(logits, targets)
                bar.update(1)
        bar.update(flush=True, linebreak=True)

        return total_loss / len(val_loader), metrics.compute()

    def _run_training_loop(
        self, train_loader: DataLoader, val_loader: DataLoader, save_dir: Path
    ):
        """运行训练循环"""
        for epoch in range(self.start_epoch, self.config.num_epochs):
            self._current_epoch = epoch  # 记录当前epoch，用于中断时显示
            epoch_str = f"Epoch {logstr.mesg(epoch + 1)}/{self.config.num_epochs}"
            logger.note(f"{'='*30} [{epoch_str}] {'='*30}")

            train_loss, train_metrics = self.train_epoch(train_loader)
            val_loss, val_metrics = self.validate(val_loader)

            # 调整学习率
            if self.use_plateau_scheduler:
                self.scheduler.step(val_metrics["avg_f1"])
            else:
                self.scheduler.step()

            # 打印结果（使用旧的 best_metrics 进行比较）
            self._log_epoch_metrics(
                epoch, train_loss, train_metrics, val_loss, val_metrics
            )

            # 更新最佳指标（在打印之后更新）
            self.best_metrics["train_loss"] = min(
                train_loss, self.best_metrics["train_loss"]
            )
            self.best_metrics["val_loss"] = min(val_loss, self.best_metrics["val_loss"])
            self.best_metrics["train_f1"] = max(
                train_metrics["avg_f1"], self.best_metrics["train_f1"]
            )
            self.best_metrics["val_f1"] = max(
                val_metrics["avg_f1"], self.best_metrics["val_f1"]
            )
            self.best_metrics["train_exact_match"] = max(
                train_metrics["exact_match_accuracy"],
                self.best_metrics["train_exact_match"],
            )
            self.best_metrics["val_exact_match"] = max(
                val_metrics["exact_match_accuracy"],
                self.best_metrics["val_exact_match"],
            )

            # 记录历史
            self.history["train_loss"].append(train_loss)
            self.history["val_loss"].append(val_loss)
            self.history["train_f1"].append(train_metrics["avg_f1"])
            self.history["val_f1"].append(val_metrics["avg_f1"])
            self.history["train_exact_match"].append(
                train_metrics["exact_match_accuracy"]
            )
            self.history["val_exact_match"].append(val_metrics["exact_match_accuracy"])
            self._log_per_key_metrics(train_metrics, val_metrics)

            # 保存最新模型（每个epoch都保存）
            self.latest_val_f1 = val_metrics["avg_f1"]
            self.latest_epoch = epoch + 1
            self._save_latest_checkpoint(
                save_dir, epoch + 1, val_metrics["avg_f1"], val_loss
            )

            # 保存最佳模型（第1个epoch之后，仅当F1提升时保存）
            if epoch == 0:
                # 第1个epoch，初始化best_val_f1但不保存
                self.best_val_f1 = val_metrics["avg_f1"]
            elif val_metrics["avg_f1"] > self.best_val_f1:
                self.best_val_f1 = val_metrics["avg_f1"]
                self._save_best_checkpoint(
                    save_dir, epoch + 1, val_metrics["avg_f1"], val_loss
                )

    def train(
        self,
        train_loader: DataLoader,
        val_loader: DataLoader,
        resume_from: Optional[str] = None,
    ) -> dict:
        """完整训练流程"""
        save_dir = Path(self.config.save_dir)
        save_dir.mkdir(parents=True, exist_ok=True)

        # 加载checkpoint（如果指定）
        if resume_from:
            if not self.load_checkpoint(resume_from):
                logger.warn("无法加载 checkpoint，从头开始训练")
                self.start_epoch = 0

        # 记录训练配置
        self._log_training_config()

        # 运行训练循环
        self._run_training_loop(train_loader, val_loader, save_dir)

        # 记录训练总结
        self._log_training_summary()

        # 保存训练历史
        # self._save_training_history(save_dir)

        return self.history

    def _save_checkpoint(
        self,
        path: Path,
        epoch: int,
        val_f1: float,
        val_loss: float,
        is_best: bool = False,
    ):
        """保存检查点"""
        checkpoint = {
            "epoch": epoch,
            "model_state_dict": self.model.state_dict(),
            "optimizer_state_dict": self.optimizer.state_dict(),
            "val_f1": val_f1,
            "val_loss": val_loss,
            "latest_val_f1": self.latest_val_f1,
            "latest_epoch": self.latest_epoch,
            "best_val_f1": self.best_val_f1,
            "history": self.history,
            "config": self.config.to_dict(),
            "is_best": is_best,
        }
        # 保存调度器状态
        if self.scheduler is not None:
            checkpoint["scheduler_state_dict"] = self.scheduler.state_dict()

        torch.save(checkpoint, path)

        # 保存同名 JSON 文件记录 checkpoint 信息
        json_path = path.with_suffix(".json")
        checkpoint_info = {
            "checkpoint_path": str(path),
            "saved_at": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "epoch": epoch,
            "val_f1": float(val_f1),
            "val_loss": float(val_loss),
            "latest_val_f1": float(self.latest_val_f1),
            "latest_epoch": self.latest_epoch,
            "best_val_f1": float(self.best_val_f1),
            "is_best": is_best,
            "current_lr": self.optimizer.param_groups[0]["lr"],
            "config": self.config.to_dict(),
            "history": self.history,
            "model_params": sum(p.numel() for p in self.model.parameters()),
        }
        with open(json_path, "w", encoding="utf-8") as f:
            json.dump(checkpoint_info, f, indent=2, ensure_ascii=False)


# ===== 预测器 ===== #


class Predictor:
    """推理预测器"""

    def __init__(
        self, model: nn.Module, config: Config = None, device: torch.device = None
    ):
        self.model = model
        self.config = config or Config()
        self.device = device or torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        self.model.to(self.device)
        self.model.eval()

    def predict(
        self,
        images: torch.Tensor,
        key_history: Optional[torch.Tensor] = None,
        threshold: float = None,
    ) -> list[dict]:
        """预测按键"""
        threshold = threshold or self.config.threshold

        # 确保输入是 5D
        if images.dim() == 4:
            images = images.unsqueeze(0)
        if key_history is not None and key_history.dim() == 2:
            key_history = key_history.unsqueeze(0)

        images = images.to(self.device)
        if key_history is not None:
            key_history = key_history.to(self.device)

        with torch.no_grad():
            logits = self.model(images, key_history)
            probs = torch.sigmoid(logits)
            preds = (probs > threshold).float()

        probs_np = probs.cpu().numpy()
        preds_np = self._postprocess(preds.cpu().numpy(), probs_np)

        results = []
        for i in range(len(preds_np)):
            results.append(
                {
                    "keys_pressed": [
                        INDEX_TO_KEY[j] for j in range(NUM_KEYS) if preds_np[i, j] > 0
                    ],
                    "probabilities": {
                        INDEX_TO_KEY[j]: float(probs_np[i, j]) for j in range(NUM_KEYS)
                    },
                    "raw_predictions": preds_np[i].tolist(),
                }
            )

        return results[0] if len(results) == 1 else results

    def _postprocess(self, preds: np.ndarray, probs: np.ndarray) -> np.ndarray:
        """后处理：解决 W+S 或 A+D 冲突"""
        preds = preds.copy()
        for i in range(len(preds)):
            if preds[i, 0] > 0 and preds[i, 2] > 0:  # W+S
                preds[i, 2 if probs[i, 0] >= probs[i, 2] else 0] = 0
            if preds[i, 1] > 0 and preds[i, 3] > 0:  # A+D
                preds[i, 3 if probs[i, 1] >= probs[i, 3] else 1] = 0
        return preds


# ===== 顶层流程编排 ===== #


class AgencyMovePipeline:
    """顶层流程编排：整合所有组件"""

    def __init__(self, config: Config = None):
        self.config = config or Config()
        self._setup_seed()

    def _setup_seed(self):
        """设置随机种子"""
        random.seed(self.config.seed)
        np.random.seed(self.config.seed)
        torch.manual_seed(self.config.seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(self.config.seed)

    @staticmethod
    def find_latest_model(save_dir: str) -> Optional[str]:
        """查找最新的模型文件"""
        save_path = Path(save_dir)
        if not save_path.exists():
            return None

        # 查找所有 .pth 文件
        model_files = list(save_path.glob("agency_move_*.pth"))
        if not model_files:
            return None

        # 按修改时间排序，返回最新的
        latest_model = max(model_files, key=lambda p: p.stat().st_mtime)
        return str(latest_model)

    @staticmethod
    def find_checkpoint_for_config(config: Config) -> Optional[str]:
        """根据配置查找匹配的checkpoint文件"""
        save_path = Path(config.save_dir)
        if not save_path.exists():
            return None

        # 生成当前配置的模型名称
        model_name = config.get_model_name()
        checkpoint_path = save_path / f"{model_name}.pth"

        if checkpoint_path.exists():
            return str(checkpoint_path)
        return None

    def train(self) -> tuple[nn.Module, dict]:
        """执行训练流程"""
        # 查找匹配的checkpoint
        resume_from = None
        if not self.config.overwrite:
            resume_from = self.find_checkpoint_for_config(self.config)
            if resume_from:
                logger.note(f"已有 checkpoint: {logstr.file(resume_from)}")
            else:
                logger.note(f"没有 checkpoint，将从头开始训练")
        else:
            logger.note("覆盖模式：忽略已有 checkpoint，从头开始训练")

        # 数据准备
        data_manager = DataManager(self.config)
        data_manager.load_all_data()
        train_loader, val_loader, test_loader = data_manager.create_data_loaders()

        # 模型创建
        model = ModelFactory.create(self.config)

        # 训练
        trainer = Trainer(self.config, model)
        history = trainer.train(train_loader, val_loader, resume_from=resume_from)

        # 测试集评估
        logger.note("在测试集上评估...")
        test_loss, test_metrics = trainer.validate(test_loader)

        logger.okay("> 测试集结果:")
        test_results = {
            "Loss": f"{test_loss:.4f}",
            "平均F1": f"{test_metrics['avg_f1']:.4f}",
            "完全匹配": f"{test_metrics['exact_match_accuracy']:.4f}",
        }
        logger.mesg(dict_to_str(test_results), indent=2)

        test_line = trainer._format_key_f1_scores(test_metrics, "测试")
        logger.mesg(test_line)

        return model, history

    def analyze(self) -> dict:
        """分析数据集"""
        data_manager = DataManager(self.config)
        return data_manager.analyze()

    def test(self, model_path: Optional[str] = None) -> dict:
        """测试模型"""
        # 如果未指定模型，查找最新的模型
        if model_path is None:
            model_path = self.find_latest_model(self.config.save_dir)
            if model_path is None:
                logger.fail(f"在 {self.config.save_dir} 中找不到任何模型文件")
                raise FileNotFoundError("没有可用的模型文件")
            logger.note(f"使用最新模型: {logstr.file(model_path)}")

        # 加载模型
        model, _ = ModelFactory.load(model_path, torch.device(self.config.device))

        # 准备数据
        data_manager = DataManager(self.config)
        data_manager.load_all_data()
        _, _, test_loader = data_manager.create_data_loaders()

        # 评估
        trainer = Trainer(self.config, model)
        test_loss, test_metrics = trainer.validate(test_loader)

        logger.okay("> 测试集结果:")
        test_results = {
            "Loss": f"{test_loss:.4f}",
            "平均F1": f"{test_metrics['avg_f1']:.4f}",
            "完全匹配": f"{test_metrics['exact_match_accuracy']:.4f}",
        }
        logger.mesg(dict_to_str(test_results), indent=2)

        test_line = trainer._format_key_f1_scores(test_metrics, "测试")
        logger.mesg(test_line)

        return test_metrics


# ===== 参数解析器 ===== #


class TrainerArgParser:
    """命令行参数解析器"""

    def __init__(self):
        self.parser = self._create_parser()

    def _create_parser(self):
        """创建参数解析器"""
        parser = argparse.ArgumentParser(description="GTAV 行为克隆模型训练")
        parser.add_argument(
            "-m",
            "--mode",
            type=str,
            default="train",
            choices=["train", "analyze", "test"],
            help="运行模式: train(训练), analyze(分析数据), test(测试模型)",
        )
        parser.add_argument(
            "-d",
            "--data-dir",
            type=str,
            default=str(DATA_DIR),
            help="数据目录路径",
        )
        parser.add_argument(
            "-p",
            "--model-path",
            type=str,
            default=None,
            help="模型文件路径（测试模式需要）",
        )
        parser.add_argument(
            "-n",
            "--history-frames",
            type=int,
            default=4,
            help="模型输入的历史参考帧数量",
        )
        parser.add_argument(
            "-b",
            "--batch-size",
            type=int,
            default=32,
            help="批次大小",
        )
        parser.add_argument(
            "-e",
            "--epochs",
            type=int,
            default=50,
            help="训练轮数",
        )
        parser.add_argument(
            "-l",
            "--lr",
            type=float,
            default=None,
            help="学习率（默认为None，使用自动调整）",
        )
        parser.add_argument(
            "-t",
            "--model-type",
            type=str,
            default="temporal",
            choices=["temporal", "simple"],
            help="模型类型: temporal(时序模型), simple(简单CNN)",
        )
        parser.add_argument(
            "-w",
            "--overwrite",
            action="store_true",
            help="覆盖模式：不加载已有 checkpoint，从头开始训练",
        )
        return parser

    def parse_args(self):
        """解析命令行参数"""
        return self.parser.parse_args()

    def create_config(self, args) -> Config:
        """从命令行参数创建配置对象"""
        return Config(
            data_dir=args.data_dir,
            history_frames=args.history_frames,
            batch_size=args.batch_size,
            num_epochs=args.epochs,
            learning_rate=args.lr,
            model_type=args.model_type,
            overwrite=args.overwrite,
        )


# ===== 主函数 ===== #


def main():
    """主函数"""
    arg_parser = TrainerArgParser()
    args = arg_parser.parse_args()
    config = arg_parser.create_config(args)

    pipeline = AgencyMovePipeline(config)

    try:
        if args.mode == "train":
            pipeline.train()
        elif args.mode == "analyze":
            pipeline.analyze()
        elif args.mode == "test":
            pipeline.test(args.model_path)
    except KeyboardInterrupt:
        print()
        logger.warn("用户中断操作 (Ctrl+C)，退出程序")


if __name__ == "__main__":
    with Runtimer():
        main()

    # Case: 训练模式：50个epoch，批次大小32
    # python -m gtaz.agency_move.train -m train -e 50 -b 32
    # python -m gtaz.agency_move.train -m train -e 50 -b 64 -w

    # Case: 分析模式：分析数据集
    # python -m gtaz.agency_move.train -m analyze

    # Case: 测试模式：使用最新模型进行测试
    # python -m gtaz.agency_move.train -m test



"""GTAV 窗口定位"""

"""
References:
- https://github.com/shibeta/JNTMbot_python/blob/main/gta_automator/game_process.py
"""


import ctypes
from ctypes import wintypes
from typing import Optional

from tclogger import TCLogger


logger = TCLogger(name="WindowLocator", use_prefix=True, use_prefix_ms=True)


# GTAV 增强版进程名
GTAV_PROCESS_NAME = "GTA5_Enhanced.exe"
# GTAV 增强版窗口标题
GTAV_WINDOW_TITLE = "Grand Theft Auto V"
# GTAV 增强版窗口类名
GTAV_WINDOW_CLASS_NAME = "sgaWindow"


# Windows API 常量
SW_RESTORE = 9
SW_SHOW = 5
SW_MINIMIZE = 6
SW_MAXIMIZE = 3
WM_CLOSE = 0x0010


class GTAVWindowLocator:
    """
    用于定位 GTAV 增强版游戏窗口的类。

    提供窗口查找、激活、前置等功能。
    """

    def __init__(
        self,
        process_name: str = GTAV_PROCESS_NAME,
        window_title: str = GTAV_WINDOW_TITLE,
        window_class_name: str = GTAV_WINDOW_CLASS_NAME,
    ):
        """
        初始化窗口定位器。

        :param process_name: 进程名称
        :param window_title: 窗口标题
        :param window_class_name: 窗口类名
        """
        self.process_name = process_name
        self.window_title = window_title
        self.window_class_name = window_class_name
        self._hwnd: Optional[int] = None

        # 加载 Windows API
        self.user32 = ctypes.windll.user32
        self.kernel32 = ctypes.windll.kernel32

    @property
    def hwnd(self) -> Optional[int]:
        """获取缓存的窗口句柄，如果无效则重新查找。"""
        if self._hwnd and self.user32.IsWindow(self._hwnd):
            return self._hwnd
        self._hwnd = self.find_window()
        return self._hwnd

    def find_window(self) -> Optional[int]:
        """
        通过窗口类名和标题查找 GTAV 窗口。

        :return: 窗口句柄 (HWND)，未找到则返回 None
        """
        hwnd = self.user32.FindWindowW(self.window_class_name, self.window_title)
        if hwnd:
            logger.okay(f"已找到 GTAV 窗口: HWND={hwnd}")
            return hwnd
        else:
            logger.warn(
                f"未找到 GTAV 窗口 (类名: {self.window_class_name}, 标题: {self.window_title})"
            )
            return None

    def is_window_valid(self) -> bool:
        """
        检查窗口是否有效（存在且可见）。

        :return: 窗口是否有效
        """
        if not self.hwnd:
            return False
        return bool(self.user32.IsWindow(self.hwnd))

    def is_window_visible(self) -> bool:
        """
        检查窗口是否可见。

        :return: 窗口是否可见
        """
        if not self.hwnd:
            return False
        return bool(self.user32.IsWindowVisible(self.hwnd))

    def is_window_foreground(self) -> bool:
        """
        检查窗口是否为前台窗口。

        :return: 窗口是否为前台窗口
        """
        if not self.hwnd:
            return False
        return self.user32.GetForegroundWindow() == self.hwnd

    def get_window_rect(self) -> Optional[tuple[int, int, int, int]]:
        """
        获取窗口的位置和大小。

        :return: (left, top, right, bottom) 或 None
        """
        if not self.hwnd:
            return None
        rect = wintypes.RECT()
        if self.user32.GetWindowRect(self.hwnd, ctypes.byref(rect)):
            return (rect.left, rect.top, rect.right, rect.bottom)
        return None

    def get_client_rect(self) -> Optional[tuple[int, int, int, int]]:
        """
        获取窗口客户区的位置和大小。

        :return: (left, top, right, bottom) 或 None
        """
        if not self.hwnd:
            return None
        rect = wintypes.RECT()
        if self.user32.GetClientRect(self.hwnd, ctypes.byref(rect)):
            # GetClientRect 返回的是相对于窗口的坐标，需要转换为屏幕坐标
            point = wintypes.POINT(0, 0)
            self.user32.ClientToScreen(self.hwnd, ctypes.byref(point))
            return (
                point.x,
                point.y,
                point.x + rect.right,
                point.y + rect.bottom,
            )
        return None

    def get_window_size(self) -> Optional[tuple[int, int]]:
        """
        获取窗口的宽度和高度。

        :return: (width, height) 或 None
        """
        rect = self.get_window_rect()
        if rect:
            return (rect[2] - rect[0], rect[3] - rect[1])
        return None

    def get_client_size(self) -> Optional[tuple[int, int]]:
        """
        获取窗口客户区的宽度和高度。

        :return: (width, height) 或 None
        """
        rect = self.get_client_rect()
        if rect:
            return (rect[2] - rect[0], rect[3] - rect[1])
        return None

    def bring_to_foreground(self) -> bool:
        """
        将窗口带到前台。

        :return: 是否成功
        """
        if not self.hwnd:
            logger.err("无法将窗口带到前台: 未找到窗口")
            return False

        try:
            # 如果窗口最小化，先恢复
            if self.user32.IsIconic(self.hwnd):
                self.user32.ShowWindow(self.hwnd, SW_RESTORE)

            # 尝试使用 SetForegroundWindow
            result = self.user32.SetForegroundWindow(self.hwnd)
            if result:
                logger.okay("成功将 GTAV 窗口带到前台")
                return True

            # 如果失败，尝试使用 AttachThreadInput 技巧
            foreground_hwnd = self.user32.GetForegroundWindow()
            if foreground_hwnd:
                foreground_thread_id = self.user32.GetWindowThreadProcessId(
                    foreground_hwnd, None
                )
                current_thread_id = self.kernel32.GetCurrentThreadId()

                if foreground_thread_id != current_thread_id:
                    self.user32.AttachThreadInput(
                        current_thread_id, foreground_thread_id, True
                    )
                    self.user32.SetForegroundWindow(self.hwnd)
                    self.user32.AttachThreadInput(
                        current_thread_id, foreground_thread_id, False
                    )

            logger.okay("成功将 GTAV 窗口带到前台")
            return True

        except Exception as e:
            logger.err(f"将窗口带到前台时出错: {e}")
            return False

    def minimize(self) -> bool:
        """
        最小化窗口。

        :return: 是否成功
        """
        if not self.hwnd:
            return False
        return bool(self.user32.ShowWindow(self.hwnd, SW_MINIMIZE))

    def maximize(self) -> bool:
        """
        最大化窗口。

        :return: 是否成功
        """
        if not self.hwnd:
            return False
        return bool(self.user32.ShowWindow(self.hwnd, SW_MAXIMIZE))

    def restore(self) -> bool:
        """
        恢复窗口（从最小化或最大化状态）。

        :return: 是否成功
        """
        if not self.hwnd:
            return False
        return bool(self.user32.ShowWindow(self.hwnd, SW_RESTORE))

    def close(self) -> bool:
        """
        关闭窗口。

        :return: 是否成功发送关闭消息
        """
        if not self.hwnd:
            logger.erro("无法关闭: 未找到 GTAV 窗口")
            return False
        result = self.user32.PostMessageW(self.hwnd, WM_CLOSE, 0, 0)
        if result:
            logger.okay("发送关闭窗口消息成功")
            return True
        else:
            logger.erro("发送关闭窗口消息失败")
            return False

    def refresh(self) -> Optional[int]:
        """
        刷新窗口句柄缓存。

        :return: 新的窗口句柄
        """
        self._hwnd = None
        return self.hwnd

    def __repr__(self) -> str:
        rect = self.get_window_rect()
        size = self.get_window_size()
        return (
            f"GTAVWindowLocator("
            f"hwnd={self._hwnd}, "
            f"valid={self.is_window_valid()}, "
            f"visible={self.is_window_visible()}, "
            f"foreground={self.is_window_foreground()}, "
            f"rect={rect}, "
            f"size={size})"
        )


def test_gtav_window_locator():
    locator = GTAVWindowLocator()

    if locator.is_window_valid():
        logger.note(f"窗口信息: {locator}")
        logger.note(f"窗口位置: {locator.get_window_rect()}")
        logger.note(f"窗口大小: {locator.get_window_size()}")
        logger.note(f"客户区位置: {locator.get_client_rect()}")
        logger.note(f"客户区大小: {locator.get_client_size()}")
        # 将窗口带到前台
        locator.bring_to_foreground()
    else:
        logger.err("GTAV 窗口未找到或无效")


if __name__ == "__main__":
    test_gtav_window_locator()

    # python -m gtaz.visions.windows



"""实时音频模板匹配检测模块

结合 sounds.py 的实时音频采集和 signals_v3.py 的模板匹配算法，
实现实时监控音频流并检测匹配事件。
"""

import argparse
import pickle
import numpy as np
import time

from pathlib import Path
from typing import Optional, Callable, Any
from tclogger import TCLogger, logstr

from .sounds import (
    SoundRecorder,
    AUDIO_DEVICE_NAME,
    SAMPLE_RATE,
    CHANNELS,
    WINDOW_MS,
    VOLUME_CHARS,
    VOLUME_BITS,
    VOLUME_GAIN,
    SAMPLES_PER_GROUP,
    VOLUME_SAMPLE_INTERVAL_MS,
)
from .signals_v3 import (
    TemplateLoader,
    AudioDataSamplesUnifier,
    FeatureExtractor,
    FeatureMatcher,
    TemplateAligner,
    TemplateFuser,
    Feature,
    MatchResult,
    MatchCandidate,
    TEMPLATES_DIR,
    TEMPLATE_REGEX,
    UNIFIED_SAMPLE_RATE,
    MATCH_WINDOW_MS,
    MATCH_STEP_MS,
    MATCH_GATE,
    CANDIDATE_MIN_OFFSET_MS,
)

logger = TCLogger(
    name="AudioDetector",
    use_prefix=True,
    use_prefix_ms=True,
    use_file=True,
    file_mode="w",
)

# 当前模块所在目录
MODULE_DIR = Path(__file__).parent
# 模板特征缓存文件
TEMPLATE_FEATURE_CACHE = MODULE_DIR / "template_feature.pkl"

# 检测间隔（毫秒）- 每隔多久执行一次匹配
DETECT_INTERVAL_MS = 200

# 绝对音量阈值（音量百分比）- 平均音量低于此值时不进行匹配
# 这是日志中显示的音量数字，基于最近采样的 RMS 值
VOLUME_THRESHOLD_PERCENT = 25


class TemplateFeatureManager:
    """模板特征管理器

    负责加载、融合和缓存模板特征。
    首次运行时计算并保存，后续运行直接加载缓存。
    """

    def __init__(
        self,
        templates_dir: Path = TEMPLATES_DIR,
        template_regex: str = TEMPLATE_REGEX,
        cache_path: Path = TEMPLATE_FEATURE_CACHE,
    ):
        self.templates_dir = templates_dir
        self.template_regex = template_regex
        self.cache_path = cache_path

        # 组件
        self.template_loader = TemplateLoader(templates_dir, template_regex)
        self.unifier = AudioDataSamplesUnifier()
        self.feature_extractor = FeatureExtractor()
        self.aligner = TemplateAligner()
        self.fuser = TemplateFuser(self.feature_extractor)

        # 融合后的模板特征
        self.fused_feature: Optional[Feature] = None

    def _compute_templates_hash(self) -> str:
        """计算模板文件的哈希值（用于检测模板是否变化）"""
        import hashlib

        self.template_loader.scan_templates()
        hash_content = ""
        for file_path in sorted(self.template_loader.template_files):
            stat = file_path.stat()
            hash_content += f"{file_path.name}:{stat.st_size}:{stat.st_mtime};"
        return hashlib.md5(hash_content.encode()).hexdigest()

    def _load_cache(self) -> Optional[Feature]:
        """从缓存加载模板特征"""
        if not self.cache_path.exists():
            return None

        try:
            with open(self.cache_path, "rb") as f:
                cache_data = pickle.load(f)

            # # 检查模板是否变化
            # cached_hash = cache_data.get("hash", "")
            # current_hash = self._compute_templates_hash()

            # if cached_hash != current_hash:
            #     logger.note("模板文件已变化，需要重新计算特征")
            #     return None

            logger.okay(f"从缓存加载模板特征: {self.cache_path}")
            return cache_data["feature"]

        except Exception as e:
            logger.warn(f"加载缓存失败: {e}")
            return None

    def _save_cache(self, feature: Feature):
        """保存模板特征到缓存"""
        try:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)
            cache_data = {
                "hash": self._compute_templates_hash(),
                "feature": feature,
            }
            with open(self.cache_path, "wb") as f:
                pickle.dump(cache_data, f)
            logger.okay(f"模板特征已缓存: {self.cache_path}")
        except Exception as e:
            logger.warn(f"保存缓存失败: {e}")

    def _compute_feature(self) -> Feature:
        """计算模板特征"""
        logger.note("计算模板特征...")

        # 扫描并加载模板
        self.template_loader.scan_templates()
        template_data = self.template_loader.load_templates()

        if len(template_data) == 0:
            raise ValueError("没有找到模板文件")

        # 统一采样率
        unified_templates = self.unifier.unify(template_data)

        # 对齐模板
        aligned_templates, offsets = self.aligner.align_templates(unified_templates)

        # 找到中心融合区域
        region_start, region_end = self.aligner.find_center_region(
            unified_templates, offsets
        )

        # 提取对齐后的区域
        extracted_templates, weights = self.aligner.extract_aligned_region(
            unified_templates, offsets, region_start, region_end
        )

        # 融合模板
        fused_feature = self.fuser.fuse_templates(extracted_templates, weights)

        logger.okay(
            f"模板特征计算完成, 维度: {fused_feature.data.shape}, "
            f"时长: {fused_feature.duration_ms:.1f}ms"
        )

        return fused_feature

    def get_feature(self, force_recompute: bool = False) -> Feature:
        """获取模板特征（优先从缓存加载）"""
        if not force_recompute:
            cached = self._load_cache()
            if cached is not None:
                self.fused_feature = cached
                return cached

        # 计算并缓存
        self.fused_feature = self._compute_feature()
        self._save_cache(self.fused_feature)
        return self.fused_feature


class AudioDetector:
    """实时音频检测器

    使用 SoundRecorder 采集音频，使用 FeatureMatcher 进行模板匹配。
    """

    def __init__(
        self,
        device_name: str = AUDIO_DEVICE_NAME,
        sample_rate: int = SAMPLE_RATE,
        channels: int = CHANNELS,
        window_ms: int = WINDOW_MS,
        gate: float = MATCH_GATE,
        detect_interval_ms: int = DETECT_INTERVAL_MS,
    ):
        self.device_name = device_name
        self.sample_rate = sample_rate
        self.channels = channels
        self.window_ms = window_ms
        self.gate = gate
        self.detect_interval_ms = detect_interval_ms

        # 模板特征管理器
        self.template_manager = TemplateFeatureManager()
        self.template_feature: Optional[Feature] = None

        # 音频录制器
        self.recorder = SoundRecorder(
            device_name=device_name,
            sample_rate=sample_rate,
            channels=channels,
            window_ms=window_ms,
        )

        # 特征匹配器（延迟初始化）
        self.matcher: Optional[FeatureMatcher] = None

        # 音频统一化器
        self.unifier = AudioDataSamplesUnifier()

        # 检测回调
        self._on_match_callback: Optional[Callable[[float, MatchResult], Any]] = None

        # 最后检测时间
        self._last_detect_time: float = 0

        # 音量显示相关
        self._sample_count = 0
        self._group_volumes: list[int] = []
        self._line_buffer: list[str] = []
        self._current_tick_detected = False  # 当前 tick 是否检测到信号

        # 组内最高分数追踪
        self._group_max_score = 0.0
        self._group_max_score_index = -1

    def _is_first_in_group(self) -> bool:
        """判断是否是组内第一个。"""
        return self._sample_count % SAMPLES_PER_GROUP == 0

    def _is_last_in_group(self) -> bool:
        """判断是否是组内最后一个。"""
        return (self._sample_count + 1) % SAMPLES_PER_GROUP == 0

    def _log_volume_char(self, volume_char: str, highlight: bool = False):
        """输出音量字符。"""
        # 只使用传入的高亮状态（当前 tick 检测到才高亮）
        display_char = logstr.okay(volume_char) if highlight else volume_char

        if self._is_first_in_group():
            self._line_buffer = []
            logger.mesg(display_char, end="")
        else:
            logger.mesg(display_char, use_prefix=False, end="")
        self._line_buffer.append(volume_char)

    @staticmethod
    def _calculate_stats(volumes: list[int]) -> tuple[int, float, int]:
        """计算音量统计信息。"""
        if not volumes:
            return 0, 0.0, 0
        return min(volumes), sum(volumes) / len(volumes), max(volumes)

    def _log_group_stats(self, score: float = None):
        """输出一组音量的统计信息（显示组内最高分数）。"""
        if self._group_volumes:
            min_vol, avg_vol, max_vol = self._calculate_stats(self._group_volumes)
            vol_strs = [
                logstr.mesg(f"{round(v):2d}") for v in [min_vol, avg_vol, max_vol]
            ]
            vol_line = "/".join(vol_strs)
            # 使用组内最高分数而非最后一次分数
            display_score = self._group_max_score
            display_index = self._group_max_score_index
            if display_score >= self.gate:
                score_str = logstr.okay(f"{display_score:.3f}✓")
                if display_index >= 0:
                    score_str += logstr.hint(f" @{display_index}")
            else:
                score_str = logstr.file(f"{display_score:.3f}")
            logger.note(f" [{vol_line}] 分数={score_str}", use_prefix=False)
        else:
            logger.note("", use_prefix=False)

    def set_on_match_callback(self, callback: Callable[[float, MatchResult], Any]):
        """设置匹配回调函数

        :param callback: 回调函数，接收 (score, match_result) 参数
        """
        self._on_match_callback = callback

    def initialize(self, force_recompute: bool = False) -> bool:
        """初始化检测器（加载模板特征）

        :param force_recompute: 是否强制重新计算模板特征
        :return: 是否成功初始化
        """
        try:
            self.template_feature = self.template_manager.get_feature(force_recompute)

            # 更新窗口大小为模板时长
            match_window_ms = self.template_feature.duration_ms

            # 输出模板特征信息
            logger.note(
                f"模板特征: duration={match_window_ms:.1f}ms, "
                f"shape={self.template_feature.data.shape}, "
                f"volume={self.template_feature.volume:.6f}"
            )

            # 创建匹配器
            self.matcher = FeatureMatcher(
                template_feature=self.template_feature,
                window_ms=match_window_ms,
                step_ms=MATCH_STEP_MS,
                gate=self.gate,
                min_offset_ms=CANDIDATE_MIN_OFFSET_MS,
                sample_rate=UNIFIED_SAMPLE_RATE,
            )

            # 计算匹配窗口采样点数
            window_samples = int(match_window_ms * UNIFIED_SAMPLE_RATE / 1000)
            logger.note(
                f"匹配参数: window_samples={window_samples}, "
                f"step_ms={MATCH_STEP_MS}, gate={self.gate}"
            )

            logger.okay("检测器初始化完成")
            return True

        except Exception as e:
            logger.warn(f"初始化失败: {e}")
            import traceback

            traceback.print_exc()
            return False

    def _preprocess_audio(self, data: np.ndarray) -> np.ndarray:
        """预处理音频数据（转单声道、统一采样率）"""
        # 转单声道
        if len(data.shape) > 1 and data.shape[1] > 1:
            data = np.mean(data, axis=1)
        elif len(data.shape) > 1:
            data = data.flatten()

        # 统一采样率
        data = self.unifier.unify_single(self.sample_rate, data)
        return data

    def detect(self, debug: bool = False) -> tuple[float, Optional[MatchResult]]:
        """执行一次检测（快速模式：只对最近的窗口进行单次匹配）

        :param debug: 是否输出调试信息
        :return: (匹配分数, 匹配结果)，无匹配时分数为 0
        """
        if self.matcher is None:
            logger.warn("检测器未初始化")
            return 0.0, None

        # 获取窗口数据
        window_data = self.recorder.buffer.get_window_data()
        if window_data is None or len(window_data) == 0:
            return 0.0, None

        # 预处理（转单声道、统一采样率）
        processed_data = self._preprocess_audio(window_data)

        # 计算需要的窗口采样点数
        window_samples = int(self.matcher.window_ms * self.matcher.sample_rate / 1000)

        # 检查数据是否足够
        if len(processed_data) < window_samples:
            if debug:
                logger.warn(f"数据长度不足: {len(processed_data)} < {window_samples}")
            return 0.0, None

        # 只取最后一个窗口的数据进行检测（快速模式）
        last_window_data = processed_data[-window_samples:]

        # 提取特征
        window_feature = self.matcher.feature_extractor.extract(last_window_data)

        # 调试：输出音量对比
        if debug:
            template_vol = self.matcher.template_feature.volume
            test_vol = window_feature.volume
            vol_ratio = test_vol / template_vol if template_vol > 1e-10 else 0
            logger.mesg(
                f"  [DEBUG] 音量对比: template={template_vol:.6f}, "
                f"test={test_vol:.6f}, ratio={vol_ratio:.4f}"
            )

        # 计算相似度
        score = self.matcher.compute_similarity(
            self.matcher.template_feature, window_feature
        )

        # 创建匹配结果
        total_duration_ms = len(processed_data) / self.matcher.sample_rate * 1000
        candidates = []
        if score >= self.matcher.gate:
            start_ms = total_duration_ms - self.matcher.window_ms
            candidates.append(
                MatchCandidate(
                    start_ms=start_ms,
                    end_ms=total_duration_ms,
                    score=score,
                    start_sample=len(processed_data) - window_samples,
                    end_sample=len(processed_data),
                )
            )

        result = MatchResult(
            test_file="",
            total_duration_ms=total_duration_ms,
            sample_rate=self.matcher.sample_rate,
            candidates=candidates,
            match_window_ms=self.matcher.window_ms,
            match_step_ms=self.matcher.step_ms,
            match_gate=self.matcher.gate,
            candidate_min_offset_ms=self.matcher.min_offset_ms,
        )

        return score, result

    def detect_sliding(
        self, return_max_raw_score: bool = True, debug: bool = False
    ) -> tuple[float, Optional[MatchResult]]:
        """执行一次检测（滑动窗口模式：扫描整个缓冲区，较慢）

        :param return_max_raw_score: 是否返回原始最高分数（即使未超过阈值）
        :param debug: 是否输出调试信息
        :return: (最高匹配分数, 匹配结果)，无匹配时分数为 0
        """
        if self.matcher is None:
            logger.warn("检测器未初始化")
            return 0.0, None

        # 获取窗口数据
        window_data = self.recorder.buffer.get_window_data()
        if window_data is None or len(window_data) == 0:
            return 0.0, None

        # 预处理
        processed_data = self._preprocess_audio(window_data)

        # 调试：输出数据信息
        if debug:
            logger.mesg(
                f"  [DEBUG] processed_data: len={len(processed_data)}, "
                f"max={np.max(np.abs(processed_data)):.6f}, "
                f"rms={np.sqrt(np.mean(processed_data**2)):.6f}"
            )

        # 执行匹配（获取原始最高分数）
        if return_max_raw_score:
            max_score, result = self._match_with_raw_score(processed_data, debug=debug)
        else:
            result = self.matcher.match(processed_data)
            # 计算最高分数
            if result.candidates:
                max_score = max(c.score for c in result.candidates)
            else:
                max_score = 0.0

        return max_score, result

    def _match_with_raw_score(
        self, test_data: np.ndarray, debug: bool = False
    ) -> tuple[float, MatchResult]:
        """匹配并返回原始最高分数（即使未超过阈值）"""
        total_duration_ms = len(test_data) / self.matcher.sample_rate * 1000

        # 计算滑动窗口参数
        window_samples = int(self.matcher.window_ms * self.matcher.sample_rate / 1000)
        step_samples = int(self.matcher.step_ms * self.matcher.sample_rate / 1000)

        # 检查数据长度是否足够
        if len(test_data) < window_samples:
            if debug:
                logger.warn(
                    f"数据长度不足: {len(test_data)} < {window_samples} (需要 {self.matcher.window_ms}ms)"
                )
            # 数据不足，返回空结果
            return 0.0, MatchResult(
                test_file="",
                total_duration_ms=total_duration_ms,
                sample_rate=self.matcher.sample_rate,
                candidates=[],
            )

        # 存储所有候选和所有分数
        all_candidates = []
        all_scores = []

        # 调试：记录第一个窗口的音量信息
        first_window_logged = False

        # 滑动窗口匹配
        start_sample = 0
        while start_sample + window_samples <= len(test_data):
            # 提取窗口特征
            window_feature = self.matcher.feature_extractor.extract_window(
                test_data, start_sample, window_samples
            )

            # 调试：输出第一个窗口的音量对比
            if debug and not first_window_logged:
                template_vol = self.matcher.template_feature.volume
                test_vol = window_feature.volume
                vol_ratio = test_vol / template_vol if template_vol > 1e-10 else 0
                logger.mesg(
                    f"  [DEBUG] 音量对比: template={template_vol:.6f}, "
                    f"test={test_vol:.6f}, ratio={vol_ratio:.4f}"
                )
                first_window_logged = True

            # 计算相似度
            score = self.matcher.compute_similarity(
                self.matcher.template_feature, window_feature
            )
            all_scores.append(score)

            # 如果超过阈值，添加候选
            if score >= self.matcher.gate:
                start_ms = start_sample / self.matcher.sample_rate * 1000
                end_ms = (
                    (start_sample + window_samples) / self.matcher.sample_rate * 1000
                )
                candidate = MatchCandidate(
                    start_ms=start_ms,
                    end_ms=end_ms,
                    score=score,
                    start_sample=start_sample,
                    end_sample=start_sample + window_samples,
                )
                all_candidates.append(candidate)

            start_sample += step_samples

        # 计算最高原始分数
        max_raw_score = max(all_scores) if all_scores else 0.0

        # 非极大值抑制
        filtered_candidates = self.matcher._non_max_suppression(all_candidates)

        result = MatchResult(
            test_file="",
            total_duration_ms=total_duration_ms,
            sample_rate=self.matcher.sample_rate,
            candidates=filtered_candidates,
            match_window_ms=self.matcher.window_ms,
            match_step_ms=self.matcher.step_ms,
            match_gate=self.matcher.gate,
            candidate_min_offset_ms=self.matcher.min_offset_ms,
        )

        return max_raw_score, result

    def detect_loop(self, duration: float = None, debug: bool = False):
        """持续检测循环

        :param duration: 检测时长（秒），None 表示持续检测
        :param debug: 是否输出调试信息（首次检测时输出）
        """
        if self.matcher is None:
            logger.warn("检测器未初始化")
            return

        # 启动音频流
        if not self.recorder.start_stream():
            logger.warn("无法启动音频流")
            return

        logger.note(
            f"开始实时检测 (阈值: {self.gate}, 间隔: {self.detect_interval_ms}ms, 音量阈值: {VOLUME_THRESHOLD_PERCENT})"
        )
        if duration:
            logger.note(f"检测时长: {duration}秒")

        start_time = time.time()
        self._sample_count = 0
        self._group_volumes = []
        self._current_tick_detected = False
        self._group_max_score = 0.0
        self._group_max_score_index = -1
        detect_count = 0
        first_detect = True

        try:
            while True:
                # 检查时长
                if duration and (time.time() - start_time) >= duration:
                    # 输出最后一组统计
                    if self._sample_count % SAMPLES_PER_GROUP != 0:
                        self._log_group_stats()
                    logger.note("检测时长已到")
                    break

                # 组开始时重置组内最高分数
                if self._is_first_in_group():
                    self._group_max_score = 0.0
                    self._group_max_score_index = -1

                # 重置当前 tick 的检测状态
                self._current_tick_detected = False

                # 检测间隔到达时执行检测（在输出音量字符之前）
                current_time = time.time()
                if (
                    current_time - self._last_detect_time
                ) * 1000 >= self.detect_interval_ms:
                    self._last_detect_time = current_time

                    # 计算最近采样的平均音量
                    recent_avg_volume = (
                        sum(self._group_volumes) / len(self._group_volumes)
                        if self._group_volumes
                        else 0
                    )

                    # 只有音量足够时才进行检测
                    if recent_avg_volume >= VOLUME_THRESHOLD_PERCENT:
                        # 首次检测时输出调试信息
                        score, result = self.detect(debug=(debug and first_detect))
                        first_detect = False
                        detect_count += 1

                        # 更新组内最高分数
                        tick_in_group = self._sample_count % SAMPLES_PER_GROUP
                        if score > self._group_max_score:
                            self._group_max_score = score
                            self._group_max_score_index = tick_in_group

                        # 如果匹配成功，设置当前 tick 检测状态并调用回调
                        if score >= self.gate:
                            self._current_tick_detected = True
                            if self._on_match_callback and result:
                                self._on_match_callback(score, result)
                    else:
                        # 音量不足，跳过检测
                        first_detect = False

                # 获取当前音量并显示（检测状态决定是否高亮）
                volume_percent = self.recorder.get_volume_percent()
                self._group_volumes.append(volume_percent)
                volume_char = self.recorder.get_volume_char(volume_percent)
                self._log_volume_char(
                    volume_char, highlight=self._current_tick_detected
                )

                # 组结束时输出统计和分数
                if self._is_last_in_group():
                    self._log_group_stats()
                    self._group_volumes = []

                self._sample_count += 1
                time.sleep(VOLUME_SAMPLE_INTERVAL_MS / 1000)

        except KeyboardInterrupt:
            # 输出最后一组统计
            if self._sample_count % SAMPLES_PER_GROUP != 0:
                self._log_group_stats()
            logger.note("\n检测到 Ctrl+C，正在退出...")

        finally:
            self.recorder.stop_stream()
            logger.okay(f"检测结束，共执行 {detect_count} 次检测")

    def detect_then_stop(
        self,
        timeout: float = None,
        on_match: Callable[[float, MatchResult], Any] = None,
    ) -> tuple[bool, float, Optional[MatchResult]]:
        """检测直到匹配成功或超时

        :param timeout: 超时时间（秒），None 表示无超时
        :param on_match: 匹配成功时的回调函数
        :return: (是否匹配成功, 最高分数, 匹配结果)
        """
        if self.matcher is None:
            logger.warn("检测器未初始化")
            return False, 0.0, None

        # 启动音频流
        if not self.recorder.start_stream():
            logger.warn("无法启动音频流")
            return False, 0.0, None

        logger.note(
            f"开始检测 (阈值: {self.gate}, 音量阈值: {VOLUME_THRESHOLD_PERCENT})，匹配到即停止"
        )
        if timeout:
            logger.note(f"超时时间: {timeout}秒")

        start_time = time.time()
        self._sample_count = 0
        self._group_volumes = []
        self._current_tick_detected = False
        self._group_max_score = 0.0
        self._group_max_score_index = -1
        matched = False
        final_score = 0.0
        final_result = None

        try:
            while True:
                # 检查超时
                if timeout and (time.time() - start_time) >= timeout:
                    # 输出最后一组统计
                    if self._sample_count % SAMPLES_PER_GROUP != 0:
                        self._log_group_stats()
                    logger.note("检测超时")
                    break

                # 组开始时重置组内最高分数
                if self._is_first_in_group():
                    self._group_max_score = 0.0
                    self._group_max_score_index = -1

                # 重置当前 tick 的检测状态
                self._current_tick_detected = False

                # 检测间隔到达时执行检测（在输出音量字符之前）
                current_time = time.time()
                if (
                    current_time - self._last_detect_time
                ) * 1000 >= self.detect_interval_ms:
                    self._last_detect_time = current_time

                    # 计算最近采样的平均音量
                    recent_avg_volume = (
                        sum(self._group_volumes) / len(self._group_volumes)
                        if self._group_volumes
                        else 0
                    )

                    # 只有音量足够时才进行检测
                    if recent_avg_volume >= VOLUME_THRESHOLD_PERCENT:
                        score, result = self.detect()

                        # 更新组内最高分数
                        tick_in_group = self._sample_count % SAMPLES_PER_GROUP
                        if score > self._group_max_score:
                            self._group_max_score = score
                            self._group_max_score_index = tick_in_group

                        # 检查是否匹配
                        if score >= self.gate:
                            self._current_tick_detected = True
                            # 先输出当前高亮的音量字符
                            volume_percent = self.recorder.get_volume_percent()
                            self._group_volumes.append(volume_percent)
                            volume_char = self.recorder.get_volume_char(volume_percent)
                            self._log_volume_char(volume_char, highlight=True)
                            # 输出当前组统计
                            if self._sample_count % SAMPLES_PER_GROUP != 0:
                                self._log_group_stats()
                            elapsed = current_time - start_time
                            logger.okay(
                                f"[{elapsed:.1f}s] 匹配成功! 分数={logstr.okay(f'{score:.4f}')}"
                            )
                            matched = True
                            final_score = score
                            final_result = result

                            # 调用回调
                            if on_match and result:
                                on_match(score, result)

                            break

                # 获取当前音量并显示
                volume_percent = self.recorder.get_volume_percent()
                self._group_volumes.append(volume_percent)
                volume_char = self.recorder.get_volume_char(volume_percent)
                self._log_volume_char(
                    volume_char, highlight=self._current_tick_detected
                )

                # 组结束时输出统计和分数
                if self._is_last_in_group():
                    self._log_group_stats()
                    self._group_volumes = []

                self._sample_count += 1
                time.sleep(VOLUME_SAMPLE_INTERVAL_MS / 1000)

        except KeyboardInterrupt:
            # 输出最后一组统计
            if self._sample_count % SAMPLES_PER_GROUP != 0:
                self._log_group_stats()
            logger.note("\n检测到 Ctrl+C，正在退出...")

        finally:
            self.recorder.stop_stream()

        return matched, final_score, final_result


class AudioDetectorArgParser:
    """命令行参数解析器"""

    def __init__(self):
        self.parser = argparse.ArgumentParser(description="实时音频模板匹配检测")
        self._add_arguments()

    def _add_arguments(self):
        self.parser.add_argument(
            "-d",
            "--duration",
            type=float,
            default=None,
            help="检测时长（秒），不指定则持续检测",
        )
        self.parser.add_argument(
            "-g",
            "--gate",
            type=float,
            default=MATCH_GATE,
            help=f"匹配阈值（默认: {MATCH_GATE}）",
        )
        self.parser.add_argument(
            "-i",
            "--interval",
            type=int,
            default=DETECT_INTERVAL_MS,
            help=f"检测间隔（毫秒，默认: {DETECT_INTERVAL_MS}）",
        )
        self.parser.add_argument(
            "-n",
            "--device-name",
            type=str,
            default=AUDIO_DEVICE_NAME,
            help=f"音频设备名称（默认: {AUDIO_DEVICE_NAME}）",
        )
        self.parser.add_argument(
            "-w",
            "--window-ms",
            type=int,
            default=WINDOW_MS,
            help=f"缓冲区窗口时长（毫秒，默认: {WINDOW_MS}）",
        )
        self.parser.add_argument(
            "-s",
            "--stop-on-match",
            action="store_true",
            help="匹配到即停止",
        )
        self.parser.add_argument(
            "-r",
            "--recompute",
            action="store_true",
            help="强制重新计算模板特征",
        )
        self.parser.add_argument(
            "--debug",
            action="store_true",
            help="输出调试信息（首次检测时显示音量对比）",
        )

    def parse(self) -> argparse.Namespace:
        return self.parser.parse_args()


def main():
    """命令行入口"""
    args = AudioDetectorArgParser().parse()

    # 创建检测器
    detector = AudioDetector(
        device_name=args.device_name,
        window_ms=args.window_ms,
        gate=args.gate,
        detect_interval_ms=args.interval,
    )

    # 初始化
    if not detector.initialize(force_recompute=args.recompute):
        logger.warn("检测器初始化失败")
        return

    # 运行检测
    if args.stop_on_match:
        matched, score, result = detector.detect_then_stop(timeout=args.duration)
        if matched:
            logger.okay(f"检测完成，匹配分数: {score:.4f}")
        else:
            logger.note("未匹配到目标音频")
    else:
        detector.detect_loop(duration=args.duration, debug=args.debug)


if __name__ == "__main__":
    main()

    # Case: 持续检测
    # python -m gtaz.audios.detects_v2

    # Case: 检测 30 秒
    # python -m gtaz.audios.detects_v2 -d 30

    # Case: 匹配到即停止
    # python -m gtaz.audios.detects_v2 -s

    # Case: 自定义阈值和间隔
    # python -m gtaz.audios.detects_v2 -g 0.5 -i 100

    # Case: 强制重新计算模板特征
    # python -m gtaz.audios.detects_v2 -r
